# Copyright (C) 2006-2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import os
from os.path import exists, join
from bsddb import dbshelve
from utils.filelock import LockFile
from OVSConstants import LOCAL_DB_HOME, POOLFS_MOUNT, LOCAL_SERVER_DB


class AgentDB(object):
    def __init__(self, db_name, db_home=LOCAL_DB_HOME, create=False):
        db_file = join(db_home, db_name)
        if not exists(db_file):
            if not create:
                raise Exception('No such db file: %s' % db_file)
            if not exists(db_home):
                os.makedirs(db_home)
            open(db_file, 'w').close()
        self.lock = LockFile(db_file)
        self.lock.acquire(10, 0.1)
        self.db = dbshelve.open(db_file)

    def get(self, key):
        return self.db.get(key)

    def put(self, key, value):
        self.db.put(key, value)

    def delete(self, key):
        self.db.delete(key)

    def has_key(self, key):
        return self.db.has_key(key)

    def keys(self):
        return self.db.keys()

    def values(self):
        return self.db.values()

    def items(self):
        return self.db.items()

    def update(self, d):
        return self.db.update(d)

    def truncate(self):
        return self.db.truncate()

    def close(self):
        try:
            self.db.close()
        finally:
            self.lock.release()


def get_cluster_db_home():
    poolfs_uuid = read_item(LOCAL_SERVER_DB, 'poolfs_uuid')
    if not poolfs_uuid:
        raise Exception('Pool filesystem unavailable.')
    return join(POOLFS_MOUNT, poolfs_uuid, 'db')


def read_item(db_name, key, db_home=LOCAL_DB_HOME):
    db = AgentDB(db_name, db_home)
    try:
        return db.get(key)
    finally:
        db.close()


def write_item(db_name, key, value, db_home=LOCAL_DB_HOME):
    db = AgentDB(db_name, db_home)
    try:
        db.put(key, value)
    finally:
        db.close()


def delete_item(db_name, key, db_home=LOCAL_DB_HOME):
    db = AgentDB(db_name, db_home)
    try:
        if db.has_key(key):
            db.delete(key)
    finally:
        db.close()


def dump_db(db_name, db_home=LOCAL_DB_HOME):
    db = AgentDB(db_name, db_home)
    try:
        return dict(db.items())
    finally:
        db.close()


def update_db(db_name, d, db_home=LOCAL_DB_HOME):
    db = AgentDB(db_name, db_home)
    try:
        db.update(d)
    finally:
        db.close()


def truncate_db(db_name, db_home=LOCAL_DB_HOME):
    db = AgentDB(db_name, db_home)
    try:
        db.truncate()
    finally:
        db.close()


def create_db(db_name, truncate=False, db_home=LOCAL_DB_HOME):
    db = AgentDB(db_name, db_home, True)
    try:
        if truncate:
            db.truncate()
    finally:
        db.close()


def delete_db(db_name, db_home=LOCAL_DB_HOME):
    db_file = join(db_home, db_name)
    if exists(db_file):
        os.unlink(db_file)
