#!/usr/bin/env python

# Copyright (C) 2006-2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import time
import sys
import socket
import os
import logging
import OVSConstants
from os.path import join, dirname, exists, getsize, basename
from SocketServer import ForkingMixIn

from utils import xmlrpclib
from utils.SimpleXMLRPCServer import SimpleXMLRPCServer, SimpleXMLRPCRequestHandler, SimpleXMLRPCDispatcher

import host
import OVSServices
from OVSLog import init_log
from OVSConstants import *
from OVSConfig import get_config, get_passwd, init_signal_handler
from OVSDaemon import daemonize
from OVSServerPool import get_pool_member_ip_list
from OVSCommons import run_cmd, expose
from password import encrypt_password


def auth(username, passwd):
    encrypted_password = encrypt_password(passwd)
    password_dict = get_passwd()
    if username not in password_dict or password_dict[username] != encrypted_password:
        raise Exception('Authorization failed: user does not exist or password error.')


def auth_ip(ip):
    if ip != "127.0.0.1":
        if ip not in get_pool_member_ip_list():
            raise Exception('Authorization failed: server not in the server pool: %s' % ip)


class AgentXMLRPCDispatcher(SimpleXMLRPCDispatcher):
    def _marshaled_dispatch(self, data, dispatch_method=None, path=None):
        params, method = xmlrpclib.loads(data)
        try:
            if dispatch_method is not None:
                response = dispatch_method(method, params)
            else:
                response = self._dispatch(method, params)
            # wrap response in a singleton tuple
            response = (response,)
            response = xmlrpclib.dumps(response, methodresponse=1, allow_none=1)
        except xmlrpclib.Fault, fault:
            response = xmlrpclib.dumps(fault, allow_none=1)
        except:
            # report exception back to server
            response = xmlrpclib.dumps(
                xmlrpclib.Fault(1, "%s:%s" % (sys.exc_type, sys.exc_value)),
                allow_none=1)
        return response


class AgentXMLRPCRequestHandler(SimpleXMLRPCRequestHandler):

    rpc_paths = ('/', '/RPC2')

    def is_rpc_path_valid(self):
        if self.rpc_paths:
            return self.path in self.rpc_paths
        else:
            # If .rpc_paths is empty, just assume all paths are legal
            return True

    def setup(self):
        config = get_config()
        if config.get("security", "ssl").lower().startswith("enable"):
            self.connection = self.request
            self.rfile = socket._fileobject(self.request, "rb", self.rbufsize)
            self.wfile = socket._fileobject(self.request, "wb", self.wbufsize)
        else:
            SimpleXMLRPCRequestHandler.setup(self)

    def do_POST(self):
        # Check that the path is legal
        if not self.is_rpc_path_valid():
            self.send_error(404, "No such page: %s" % self.path)
            return

        try:
            src_ip = self.client_address[0]
            OVSConstants.MANAGER_SRC_IP = src_ip
            if "authorization" in self.headers:
                username, password = self.headers["authorization"].split(" ")[1].decode("base64").split(":")
                auth(username, password)
            else:
                auth_ip(src_ip)
        except:
            errmsg = "Unauthorized access attempt from %s!" % (self.client_address,)
            logging.exception(errmsg)
            self.send_error(403, errmsg)
            return

        config = get_config()
        if config.get("security", "ssl").lower().startswith("enable"):
            try:
                # get arguments
                data = self.rfile.read(int(self.headers["content-length"]))
                # In previous versions of SimpleXMLRPCServer, _dispatch
                # could be overridden in this class, instead of in
                # SimpleXMLRPCDispatcher. To maintain backwards compatibility,
                # check to see if a subclass implements _dispatch and dispatch
                # using that method if present.
                response = self.server._marshaled_dispatch(
                        data, getattr(self, '_dispatch', None), self.path
                    )
            except: # This should only happen if the module is buggy
                # internal error, report as HTTP server error
                self.send_response(500)
                self.end_headers()
            else:
                # got a valid XML RPC response
                self.send_response(200)
                self.send_header("Content-type", "text/xml")
                self.send_header("Content-length", str(len(response)))
                self.end_headers()
                self.wfile.write(response)

                # shut down the connection
                self.wfile.flush()
                self.connection.shutdown()
        else:
            try:
                # get arguments
                data = self.rfile.read(int(self.headers["content-length"]))
                # In previous versions of SimpleXMLRPCServer, _dispatch
                # could be overridden in this class, instead of in
                # SimpleXMLRPCDispatcher. To maintain backwards compatibility,
                # check to see if a subclass implements _dispatch and dispatch
                # using that method if present.
                response = self.server._marshaled_dispatch(
                        data, getattr(self, '_dispatch', None), self.path
                    )
            except: # This should only happen if the module is buggy
                # internal error, report as HTTP server error
                self.send_response(500)
                self.end_headers()
            else:
                # got a valid XML RPC response
                self.send_response(200)
                self.send_header("Content-type", "text/xml")
                self.send_header("Content-length", str(len(response)))
                self.end_headers()
                self.wfile.write(response)

                # shut down the connection
                self.wfile.flush()
                self.connection.shutdown(1)

    def log_message(self, format, *args):
        logging.info(format % args)


class OVSAgentServer(ForkingMixIn, SimpleXMLRPCServer):

    allow_reuse_address = True
    max_children = 500 # default 40 in ForkingMixin

    def __init__(self, addr, requestHandler=AgentXMLRPCRequestHandler, logRequests=1):
        SimpleXMLRPCServer.__init__(self, addr, requestHandler, logRequests)
        config = get_config()
        if config.get("security", "ssl").lower().startswith("enable"):
            from OpenSSL import SSL
            KEYFILE = join(CONFIG_DIR, "cert/key.pem")
            CERTFILE = join(CONFIG_DIR, "cert/certificate.pem")
            ctx = SSL.Context(SSL.SSLv23_METHOD)
            ctx.set_options(SSL.OP_NO_SSLv2)
            ctx.use_privatekey_file(KEYFILE)
            ctx.use_certificate_file(CERTFILE)
            self.socket = SSL.Connection(ctx, socket.socket(self.address_family, self.socket_type))
            self.server_bind()
            self.server_activate()

        api_v1_dispatcher = AgentXMLRPCDispatcher()

        for (api, func) in OVSServices.api_v1.items():
            api_v1_dispatcher.register_function(expose(api)(func), api)

        self.dispatchers = {
            '/': api_v1_dispatcher,
            '/RPC2': api_v1_dispatcher,
            '/api/1': api_v1_dispatcher}

        requestHandler.rpc_paths = ['/', '/RPC2', '/api/1']

    def _marshaled_dispatch(self, data, dispatch_method=None, path=None):
        try:
            response = self.dispatchers[path]._marshaled_dispatch(
               data, dispatch_method, path)
        except:
            # report low level exception back to server
            # (each dispatcher should have handled their own
            # exceptions)
            response = xmlrpclib.dumps(
                xmlrpclib.Fault(1, "%s:%s" % (sys.exc_type, sys.exc_value)),
                allow_none=1)
        return response

    def verify_request(self, request, client_address):
        """
        overwrite original verify_request to implement IP based access control
        """
        #--------------------<> internal function <>----
        def match(rule, ip):
            #check if ip is in the range
            try:
                #e.g.: 219.142.73.80
                if ip == rule: return True
                #e.g.: 219.142.73.*
                if rule.endswith("*") and ip.startswith(rule[:-1]): return True
                #e.g.: 219.142.73.0/20
                vip, bits = rule.split("/")
                vip_str = "".join(["%02x" % int(x) for x in vip.split(".")])
                ip_str = "".join(["%02x" % int(x) for x in ip.split(".")])
                q, r = divmod(int(bits), 4)
                if vip_str[:q] == ip_str[:q]:
                    if int(vip_str[q], 16) >> (4 - r) == int(ip_str[q], 16) >> (4 - r):
                        return True
            except: #in case there are errors in rules
                pass
        #--------------------<> main processing <>----
        #check ip access control rules, sequence: allow=>deny
        (ip, port) = client_address

        # always allow connect from localhost
        if ip == '127.0.0.1': return True

        config = get_config()
        #allow rules
        allow = False
        allow_rules = config.get("security", "allow").split(",")
        for rule in allow_rules:
            if match(rule.strip(), ip):
                allow = True
                break
        if not allow: return False

        #deny rules
        deny = False
        deny_rules = config.get("security", "deny").split(",")
        for rule in deny_rules:
            if match(rule.strip(), ip):
                deny = True
                break
        if deny: return False

        #nothing wrong, finally
        return True


def start_server():
    try:
        config = get_config()
        server = OVSAgentServer(('', int(config.get("agent", "port"))),
                                AgentXMLRPCRequestHandler, logRequests=False)
        logging.info("OVSAgentServer start serving!")
        logging.info("Oracle VM Server version: %s", host.get_ovs_version())
        try:
            server.serve_forever()
        except KeyboardInterrupt, e: #Ctrl+C
            logging.info("OVSAgentServer: server stopped! => %s" % (e,))
            server.server_close()
            sys.exit(0)
    except Exception, e:
        logging.error("OVSAgentServer.start_server(): error. => %s" % (e,))
        sys.exit(1)


def main(front):
    pid_filename = basename(__file__).split(".")[0]
    if not front:
        daemonize(pid_filename)
    init_signal_handler()
    init_log()

    try:
        run_cmd(['xm', 'sched-credit', '-d', '0', '-w', '65535'])
    except:
        pass

    start_server()


if __name__ == "__main__":
    front = False
    if len(sys.argv) > 1:
        if sys.argv[1] == 'front':
            front = True
    main(front)
