# Copyright (C) 2006-2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import os
import re
import sys
import string
import signal
import logging
from OVSConstants import LOCAL_APROC_DB
from OVSAgentDB import read_item, write_item, delete_item, dump_db
from OVSNotificationProxy import notify_manager


def clean_aproc_db():
    for (key, pid) in dump_db(LOCAL_APROC_DB).items():
        if not os.path.exists("/proc/%s" % pid):
            delete_item(LOCAL_APROC_DB, key)


def init_async_procenv():
    clean_aproc_db()


def get_proc_id(pid):
    return read_item(LOCAL_APROC_DB, str(pid))


def remove_proc_id(pid):
    delete_item(LOCAL_APROC_DB, str(pid))


def ovs_async_proc(func, *args):
    """
    Runs func as a daemon, add it PID to the process DB table (aproc)

    @param func    The function to run
    @param args    A list of arguments to pass to the function

    @return        If successful, returns the process ID.
                   Raises an exception on failure.

    Restriction:
        The function must be known to OVSServices, otherwise this call will fail.
    """
    if func == None or func == '':
        raise Exception ("ovs_async_proc: NULL function name")

    clean_aproc_db()

    from OVSServices import api

    func_addr = api.get(func)
    if func_addr == None:
        raise Exception('ovs_async_proc: function: %s not found in OVSServices' % func)

    logging.debug('async call fuction: %s', func)

    # Create the deamon
    try:
        pid = os.fork()
        if pid > 0:
            #exit first parent
            os._exit(0)
    except OSError:
        raise Exception('ovs_async_proc: failed to fork %s %s' % (func, str(args)))

    # decouple from parent
    os.chdir("/")
    os.setsid()
    os.umask(022)

    # second fork
    try:
        pid = os.fork()
        if pid > 0:
            write_item(LOCAL_APROC_DB, str(pid), pid)
            return pid
    except OSError:
        raise Exception('ovs_async_proc: failed to fork %s %s' % (func, str(args)))

    pid = os.getpid()
    status = 0
    try:
        output = func_addr(*args)
    except Exception, err:
        status = 1
        output = str(err)

    delete_item(LOCAL_APROC_DB, str(pid))

    if output is None:
        output = ''

    logging.debug('ovs_async_proc: pid %s status %s output %s', pid, status, output)
    try:
        notify_manager('{ASYNC_PROC} exit PID %s' % pid, {'status': status, 'output': output})
    except Exception, err:
        logging.warn('Error sending notification: pid(%s) status(%s) output(%s) error(%s)' % (pid, status, output, err))


def ovs_async_proc_status(pid):
    """
    Gets the running status of a daemon as defined by the Linux kernel

    @param pid     Process ID

    @return        If successful, returns procesess running status
                   Raises an exception on failure.
    Restriction:
       - PID must exists in the process DB table (aproc)
    """
    if pid == None or pid == '':
        raise Exception('ovs_async_proc_status: PID is NULL')

    id = get_proc_id(pid)
    if id == None:
        raise Exception('ovs_async_proc_status: PID %d does not exist in DB' % int(pid))

    if not os.path.exists("/proc/%s"  % pid):
        raise Exception('ovs_async_proc_status: Process %d does not exist' % int(pid))

    f = open('/proc/%d/status' % int(pid),'r')
    try:
        lines = f.readlines()
    finally:
        f.close()
    for ln in lines:
        if ln.startswith( 'State:' ):
            junk, state, text = ln.split( None, 2 )
            return "Process %d  state: %s %s" % (int(pid), state, text)


def ovs_async_proc_signal(sig, pid):
    """
    Sends SIGKILL, SIGUSR1 to a running process

    @param sig     (kill | stop)
                   kill: sends a SIGKILL to the process
                   stop: sends a SIGUSR1 to the process
    @param pid     Process ID to signal

    @return        If successful, returns procesess ID, and signal delivered
                   Raises an exception on failure.

    Restriction:
       - For a process to receive SIGUSR1, a processes must register a
         signal handler. The purpose of this signal is to tell the process
         to clean up and exit vs just killing it.
       - PID must exists in the process DB table (aproc)
    """
    if sig != "stop" and sig != "kill":
        raise Exception("ovs_async_proc_signa: signal %s is not supported" % sig)

    id = get_proc_id(pid)
    if id == None:
        raise Exception('ovs_async_proc_signal: PID %d does not exist in DB' % int(pid))

    if not os.path.exists("/proc/%s"  % pid):
        raise Exception('ovs_async_proc_status: Process %d does not exist' % int(pid))

    if sig == "kill":
        try:
            os.kill(int(pid), signal.SIGKILL)
            remove_proc_id(pid)
            return "Process %d is killed" % int(pid)
        except OSError, e:
            raise Exception('ovs_async_proc_signal: failed to send sig %s to proc %d => %s' % (sig, int(pid), str(e)))
    else:
        try:
            os.kill(int(pid), signal.SIGUSR1)
            #In order to prevent the DB from getting out of sync, do not remove the PID
            #at this point, as the function may not respond to SIGUSR1 and continue to run.
            #the PID will get cleaned up the next ovs_async_proc() gets called.
            #remove_proc_id(pid)
            return "Process %d is stopped" % int(pid)
        except OSError, e:
            raise Exception('ovs_async_proc_signal: failed to send sig %s to proc %d => %s' % (sig, int(pid), str(e)))


def async_proc_prog(status):
    notification = '{ASYNC_PROC} progress PID %s' % (os.getpid(),)
    logging.debug('async_proc_prog: notification %s status %s', notification, status)
    try:
        notify_manager(notification, status)
    except Exception, err:
        logging.warn('Error sending notification: %s status: %s error: %s' % (
                     notification, status, err))


def async_proc_cache(cache):
    notification = '{ASYNC_PROC} cache PID %s' % (os.getpid(),)
    try:
        notify_manager(notification, cache)
    except Exception, err:
        logging.warn('Error sending notification: %s cache: %s error: %s' % (
                     notification, cache, err))
