# Copyright (C) 2006-2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

from os import makedirs, remove, listdir
from os.path import join, exists
import time
import string
import logging
import copy

from OVSCommons import run_cmd
from OVSStat import reload_monitor
from OVSConstants import *
from OVSOCFS2 import ocfs2_file_read
from OVSAgentDB import (read_item, write_item, delete_item, dump_db,
        truncate_db, create_db, update_db, get_cluster_db_home)
from OVSServerPool import get_poolfs_uuid, get_membership_state, get_ocfs2_cluster_name
from OVSPoolFileSystem import (create_dmnfs_device, remove_dmnfs_device,
        mount_device, umount_device, parse_metafile, rename_file)
from utils import filelock
from cluster import Cluster


def is_cluster_online():
    cmd = ["o2cb", "cluster-status"]
    try:
        # returns 0 if cluster online, 1 otherwise
        run_cmd(cmd)
        return True
    except:
        return False


def server_cluster_debug():
    d = {}
    debug_paths = [
            "/sys/kernel/debug/o2dlm",
            "/sys/kernel/debug/o2hb",
            "/sys/kernel/debug/o2net",
            "/sys/kernel/debug/ocfs2",]
    for p in debug_paths:
        if exists(p):
            d[p] = listdir(p)
    cmd = ["service", "o2cb", "status"]
    d[" ".join(cmd)] = run_cmd(cmd)
    logging.debug("cluster debug: %s", d)


def server_cluster_online():
    """
    Register nodes and start heartbeating.

    @return     None. Raises an exception on failure.
    """
    server_cluster_debug()

    lock = filelock.FileLock(CLUSTER_LOCK)
    lock.acquire(wait=60)
    try:
        cmd = ["service", "o2cb", "online"]
        run_cmd(cmd)
    finally:
        lock.release()

    server_cluster_debug()

    if not is_cluster_online():
        raise Exception("Unable to online cluster.")


def server_cluster_offline():
    """
    Stop heatbeating and unregister nodes.

    @return     None. Raises an exception on failure.
    """
    server_cluster_debug()

    lock = filelock.FileLock(CLUSTER_LOCK)
    lock.acquire(wait=60)
    try:
        cmd = ["rm", "-fr", DLM_DOMAIN]
        run_cmd(cmd)
        cmd = ["service", "o2cb", "offline"]
        run_cmd(cmd)
    finally:
        lock.release()

    server_cluster_debug()

    if is_cluster_online():
        raise Exception("Unable to offline cluster.")


def update_config(config, update, default={}):
    """
    Updates the configuration with the given update.
    """
    d = {}
    for ln in config:
        ln = ln.split('#')[0].strip()
        if ln:
            parts = map(string.strip, ln.split('='))
            if len(parts) == 2 and parts[0]:
                d[parts[0]] = parts[1]

    conf = copy.copy(default)
    conf.update(d)
    conf.update(update)

    new_config = []
    for k, v in conf.items():
        line = '%s=%s\n' % (k, v)
        new_config.append(line)

    return new_config


def file_read(path):
    f = open(path, "r")
    try:
        return f.read()
    finally:
        f.close()


def file_write(path, string):
    f = open(path, "w")
    try:
        f.write(string)
    finally:
        f.close()


def file_readlines(path):
    f = open(path, "r")
    try:
        return f.readlines()
    finally:
        f.close()


def file_writelines(path, lines):
    f = open(path, "w")
    try:
        f.writelines(lines)
    finally:
        f.close()


def parse_cluster_layout(cfg_file=CLUSTER_CONF):
    """
    @return:    {'node':[<dict of node>],
                 'cluster': [<dict of cluster>],
                 'heartbeat': [<dict of heartbeat>],}
    e.g.,
    {'cluster': [{'heartbeat_mode': 'global',
                  'node_count': '3',
                  'name': 'ocfs2test'}],
     'node': [{'ip_port': '7777',
               'cluster': 'ocfs2test',
               'ip_address': '10.149.76.185',
               'number': '0',
               'name': 'ovs085'},
              {'ip_port': '7777',
               'cluster': 'ocfs2test',
               'ip_address': '10.149.76.186',
               'number': '1',
               'name': 'ovs086'},
              {'ip_port': '7777',
               'cluster': 'ocfs2test',
               'ip_address': '10.149.76.187',
               'number': '2',
               'name': 'ovs087'}],
     'heartbeat': [{'cluster': 'ocfs2test',
                    'region': '380832D5B2A64C30948744568F28B302'}}
    """
    conf_dict = {'node': [], 'cluster': [], 'heartbeat': []}
    for line in file_readlines(cfg_file):
        line = line.split('#')[0].strip()
        if not line:
            continue
        elif line.startswith('node:'):
            d = {}
            conf_dict['node'].append(d)
        elif line.startswith('cluster:'):
            d = {}
            conf_dict['cluster'].append(d)
        elif line.startswith('heartbeat:'):
            d = {}
            conf_dict['heartbeat'].append(d)
        else:
            parts = map(string.strip, line.split('='))
            if len(parts) == 2 and parts[0]:
                d[parts[0]] = parts[1]
    return conf_dict


def config_cluster_layout(cluster_conf, cfg_file=CLUSTER_CONF):
    if exists(cfg_file):
        cluster_layout = parse_cluster_layout(cfg_file)
    else:
        cluster_layout = {'cluster': [], 'node': [], 'heartbeat': []}

    cluster = Cluster(cfg_file)

    for cl in cluster_layout['cluster']:
        if cl not in cluster_conf['cluster']:
            cluster.remove_cluster(cl)
    for cl in cluster_conf['cluster']:
        if cl not in cluster_layout['cluster']:
            cluster.add_cluster(cl)
            cluster.heartbeat_mode(cl)

    for nd in cluster_layout['node']:
        if nd not in cluster_conf['node']:
            cluster.remove_node(nd)
    for nd in cluster_conf['node']:
        if nd not in cluster_layout['node']:
            cluster.add_node(nd)

    #for hb in cluster_layout['heartbeat']:
    #    if hb not in cluster_conf['heartbeat']:
    #        cluster.remove_heartbeat(hb)
    for hb in cluster_conf['heartbeat']:
        if hb not in cluster_layout['heartbeat']:
            cluster.add_heartbeat(hb)


def configure_server_for_cluster(o2cb_conf, cluster_conf,
            poolfs_type, poolfs_target, poolfs_uuid, poolfs_nfsbase_uuid):
    """
    Configures the server for the cluster.

    @param      o2cb_conf - dict of o2cb sysconfig
    @param      cluster_conf - cluster configuration file
    @param      poolfs_type - "nfs" or "lun"
    @param      poolfs_target -
                physical location of the pool filesystem
    @param      poolfs_uuid -
                a unique identifier for the pool filesystem
    @param      poolfs_nfsbase_uuid -
                a unique identifier for the base NFS filesystem

    @return     None. Raises an exception on failure.
    """
    if poolfs_type not in ("nfs", "lun"):
        raise ValueError("Invalid pool filesystem type: %s" % poolfs_type)

    # check for membership state
    state = get_membership_state()
    if state != MEMBERSHIP_STATE_POOLED:
        raise Exception("Cannot configure server for cluster "
                                "in membership state %s" % state)

    # check for cluster.conf and cluster layout in configfs
    if exists(CLUSTER_CONF):
        raise Exception("/etc/ocfs2/cluster.conf already exists.")

    cluster_name = get_ocfs2_cluster_name()
    if cluster_name:
        raise Exception("Already in cluster: %s" % cluster_name)

    # check for poolfs uuid
    if poolfs_type == "nfs":
        # mount poolfs nfsbase
        mount_point = join(NFS_MOUNT, poolfs_nfsbase_uuid)
        mount_device(poolfs_target, mount_point)

        # read metadata
        poolfs_meta = join(mount_point, POOLFS_META)
        contents = file_read(poolfs_meta)

    else:
        contents = ocfs2_file_read(poolfs_target, POOLFS_META)

    d = parse_metafile(contents)
    ovs_poolfs_uuid = d.get("OVS_POOLFS_UUID")
    if not ovs_poolfs_uuid:
        raise Exception("No pool filesystem UUID in the metafile.")
    if ovs_poolfs_uuid != poolfs_uuid:
        raise Exception("Pool filesystem UUID does not match "
                                        "the metafile record.")

    if not exists(CLUSTER_CONF_DIR):
        makedirs(CLUSTER_CONF_DIR)

    if poolfs_type == "lun":
        poolfs_device = poolfs_target
    else:
        poolfs_device = join("/dev/mapper/", POOLFS_DMNFS_DEVICE)

    stage = 0

    try:
        # create dmnfs device
        if poolfs_type == "nfs":
            poolfs_img = join(mount_point, POOLFS_IMG)
            create_dmnfs_device(poolfs_img, POOLFS_DMNFS_DEVICE)

        # update o2cb config
        config = file_readlines(O2CB_CONF)
        default_conf = {
            "O2CB_ENABLED": "true",
            "O2CB_STACK": "o2cb",
            "O2CB_BOOTCLUSTER": "ocfs2",
            "O2CB_HEARTBEAT_THRESHOLD": "31",
            "O2CB_IDLE_TIMEOUT_MS": "60000",
            "O2CB_KEEPALIVE_DELAY_MS": "2000",
            "O2CB_RECONNECT_DELAY_MS": "2000",
            }
        new_config = update_config(config, o2cb_conf, default_conf)
        file_writelines(O2CB_CONF, new_config)
        stage = 1

        # write cluster.conf
        if isinstance(cluster_conf, dict):
            config_cluster_layout(cluster_conf)
        else:
            file_write(CLUSTER_CONF, cluster_conf)
        stage = 2

        # cluster online
        server_cluster_online()

        # mount poolfs
        poolfs_mount_point = join(POOLFS_MOUNT, poolfs_uuid)
        mount_device(poolfs_device, poolfs_mount_point)

        # update db
        poolfs_info = {
            "poolfs_type": poolfs_type,
            "poolfs_target": poolfs_target,
            "poolfs_uuid": poolfs_uuid,
            "poolfs_nfsbase_uuid": poolfs_nfsbase_uuid}
        update_db(LOCAL_SERVER_DB, poolfs_info)

    except Exception, e:
        try:
            # umount poolfs
            umount_device(poolfs_device)

            # cluster offline
            server_cluster_offline()

            # remove cluster.conf
            if stage > 1:
                if exists(CLUSTER_CONF):
                    remove(CLUSTER_CONF)

            # restore o2cb config
            if stage > 0:
                file_writelines(O2CB_CONF, config)

            if poolfs_type == "nfs":
                # remove dmnfs device
                remove_dmnfs_device(POOLFS_DMNFS_DEVICE)
                # umount poolfs nfsbase
                umount_device(poolfs_target)

        finally:
            raise e


def deconfigure_server_for_cluster(poolfs_uuid):
    """
    Removes the cluster configuration.

    @param      poolfs_uuid -
                a unique identifier for the pool filesystem

    @return     None. Raises an exception on failure.
    """
    # check for membership state
    state = get_membership_state()
    if state != MEMBERSHIP_STATE_HB_CONFIGURED:
        raise Exception("Cannot deconfigure server for cluster "
                                    "in membership state %s" % state)

    # check for poolfs uuid
    uuid = get_poolfs_uuid()

    if not uuid:
        raise Exception("There is no db record of pool filesystem UUID.")

    if poolfs_uuid != uuid:
        raise Exception("Pool filesystem UUID does not match "
                                            "the db record.")

    # check for cluster.conf and cluster layout in configfs
    if not exists(CLUSTER_CONF):
        raise Exception("/etc/ocfs2/cluster.conf does not exist.")

    d = dump_db(LOCAL_SERVER_DB)
    poolfs_type = d.get("poolfs_type")
    poolfs_target = d.get("poolfs_target")
    poolfs_nfsbase_uuid = d.get("poolfs_nfsbase_uuid")
    poolfs_uuid = d.get("poolfs_uuid")

    if poolfs_type == "lun":
        poolfs_device = poolfs_target
    else:
        poolfs_device = join("/dev/mapper/", POOLFS_DMNFS_DEVICE)

    stage = 0

    try:
        # umount poolfs
        umount_device(poolfs_device)

        # cluster offline
        server_cluster_offline()

        if poolfs_type == "nfs":
            # remove dmnfs device
            remove_dmnfs_device(POOLFS_DMNFS_DEVICE)
            # umount poolfs nfsbase
            umount_device(poolfs_target)

        # backup cluster.conf
        rename_file(CLUSTER_CONF, CLUSTER_CONF_ORIG)
        stage = 1

        # delete db records
        poolfs_info = {
            "poolfs_type": None,
            "poolfs_target": None,
            "poolfs_uuid": None,
            "poolfs_nfsbase_uuid": None}
        update_db(LOCAL_SERVER_DB, poolfs_info)

    except Exception, e:
        try:
            # restore cluster.conf
            if stage > 0:
                rename_file(CLUSTER_CONF_ORIG, CLUSTER_CONF)

            if poolfs_type == "nfs":
                # mount poolfs nfsbase
                mount_point = join(NFS_MOUNT, poolfs_nfsbase_uuid)
                mount_device(poolfs_target, mount_point)
                # recreate dmnfs device
                poolfs_img = join(mount_point, POOLFS_IMG)
                create_dmnfs_device(poolfs_img, POOLFS_DMNFS_DEVICE)

            # cluster online
            server_cluster_online()

            # mount poolfs
            poolfs_mount_point = join(POOLFS_MOUNT, poolfs_uuid)
            mount_device(poolfs_device, poolfs_mount_point)

        finally:
            raise e


def update_cluster_configuration(cluster_conf):
    """
    Updates /etc/ocfs2/cluster.conf with cluster_conf.

    @param      cluster_conf - cluster configuration file

    @return     None. Raises an exception on failure.
    """
    # check for membership state
    state = get_membership_state()
    if state not in \
        (MEMBERSHIP_STATE_HB_CONFIGURED, MEMBERSHIP_STATE_CLUSTERED):
        raise Exception("Cannot update cluster configuration "
                            "in membership state %s" % state)

    stage = 0

    try:
        # backup cluster.conf
        rename_file(CLUSTER_CONF, CLUSTER_CONF_ORIG)
        stage = 1

        # update cluster.conf
        if isinstance(cluster_conf, dict):
            config_cluster_layout(cluster_conf)
        else:
            file_write(CLUSTER_CONF, cluster_conf)

        # cluster online
        server_cluster_online()
    except Exception, e:
        try:
            # restore cluster.conf
            if stage > 0:
                rename_file(CLUSTER_CONF_ORIG, CLUSTER_CONF)
            # cluster online
            server_cluster_online()
        finally:
            raise e


def cluster_add_server(poolfs_uuid, init=False):
    """
    Adds a new server to the cluster.

    @param      poolfs_uuid - pool filesystem uuid
    @param      init - True if it is the first cluster member

    @return     None. Raises an exception on failure.
    """
    # check for membership state
    state = get_membership_state()
    if state == MEMBERSHIP_STATE_CLUSTERED:
        return
    elif state != MEMBERSHIP_STATE_HB_CONFIGURED:
        raise Exception("Cannot add server to cluster "
                        "in membership state %s" % state)

    # get db dict
    d = dump_db(LOCAL_SERVER_DB)

    # check for poolfs uuid
    if poolfs_uuid != d.get("poolfs_uuid"):
        raise Exception("Pool filesystem UUID does not match "
                                            "the db record.")
    # hostname and server info
    hostname = d.get("registered_hostname")
    srv = {
        "roles": d.get("roles"),
        "node_number": d.get("node_number"),
        "registered_ip": d.get("registered_ip"),
        "is_master": d.get("is_master")}

    # validate server roles
    master_flag = d.get("is_master")
    if init and not master_flag:
        raise Exception("Initial server must be server pool master.")
    if not init and master_flag:
        raise Exception("Server pool master must be initial server.")

    try:
        # setup qlayer-monitor
        monitor_base_dir_clustered = join(POOLFS_MOUNT, poolfs_uuid, "qlayer")
        reload_monitor(monitor_base_dir_clustered)

        db_home = get_cluster_db_home()

        if init:
            # shared db initialization
            create_db("server_pool_servers", True, db_home)
            write_item("server_pool_servers", hostname, srv, db_home)
            create_db("server_pool", True, db_home)
            pool_info = {
                "pool_alias": d.get("pool_alias"),
                "pool_uuid": d.get("pool_uuid"),
                "pool_virtual_ip": d.get("pool_virtual_ip"),
                "pool_master_hostname": hostname,
                "pool_member_ip_list": d.get("pool_member_ip_list"),
                "auto_remaster": True}
            update_db("server_pool", pool_info, db_home)
            create_db("monitored_vms", True, db_home)
        else:
            # update server information
            write_item("server_pool_servers", hostname, srv, db_home)

        # remove pool_member_ip_list from local db
        if "pool_member_ip_list" in d:
            delete_item(LOCAL_SERVER_DB, "pool_member_ip_list")

        # mark server as clustered
        write_item(LOCAL_SERVER_DB, "clustered", True)

    except Exception, e:
        try:
            # setup qlayer-monitor
            reload_monitor()
        finally:
            raise e


def create_cluster(poolfs_uuid):
    """
    Creates a new cluster containing the pool master.

    @param      poolfs_uuid - pool filesystem uuid

    @return     None. Raises an exception on failure.
    """
    cluster_add_server(poolfs_uuid, True)


def join_cluster(poolfs_uuid):
    """
    Adds an additional server to the cluster.

    @param      poolfs_uuid - pool filesystem uuid

    @return     None. Raises an exception on failure.
    """
    cluster_add_server(poolfs_uuid, False)


def cluster_remove_server(poolfs_uuid, final=False):
    """
    Removes a server from the cluster.

    @param      poolfs_uuid - pool filesystem uuid
    @param      final - True if it is the final server

    @return     None. Raises an exception on failure.
    """
    # check for membership state
    state = get_membership_state()
    if state == MEMBERSHIP_STATE_HB_CONFIGURED:
        return
    elif state != MEMBERSHIP_STATE_CLUSTERED:
        raise Exception("Cannot remove server from cluster "
                        "in membership state %s" % state)

    # check for poolfs uuid
    uuid = get_poolfs_uuid()
    if poolfs_uuid != uuid:
        raise Exception("Pool filesystem UUID does not match "
                                            "the db record.")
    # get db dict
    d = dump_db(LOCAL_SERVER_DB)

    # validate server role
    master_flag = d.get("is_master")
    if final and not master_flag:
        raise Exception("Final server in the cluster must be the pool master.")
    if not final and master_flag:
        raise Exception("Pool master cannot be removed from the cluster.")

    db_home = get_cluster_db_home()

    # update local db
    ip_list = read_item("server_pool", "pool_member_ip_list", db_home)
    update = {
        "clustered": False,
        "pool_member_ip_list": ip_list}
    update_db(LOCAL_SERVER_DB, update)

    try:
        # setup qlayer-monitor
        reload_monitor()
        if final:
            # wait before truncating db
            time.sleep(5)
            # truncate db
            truncate_db("server_pool", db_home)
            truncate_db("server_pool_servers", db_home)
        else:
            # remove db record
            delete_item("server_pool_servers", d.get("registered_hostname"), db_home)
    except Exception, e:
        try:
            # restore local db
            update = {
                "clustered": True,
                "pool_member_ip_list": None}
            update_db(LOCAL_SERVER_DB, update)
            #setup qlayer-monitor
            monitor_base_dir_clustered = join(POOLFS_MOUNT, poolfs_uuid, "qlayer")
            reload_monitor(monitor_base_dir_clustered)
        finally:
            raise e


def destroy_cluster(poolfs_uuid):
    """
    Removes the single, final server from the cluster.

    @param      poolfs_uuid - pool filesystem uuid

    @return     None. Raises an exception on failure.
    """
    cluster_remove_server(poolfs_uuid, True)


def leave_cluster(poolfs_uuid):
    """
    Removes a server from the cluster.

    @param      poolfs_uuid - pool filesystem uuid

    @return     None. Raises an exception on failure.
    """
    cluster_remove_server(poolfs_uuid, False)
