# Copyright (C) 2006-2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import logging
import random
import os
import subprocess
import copy
import time
from xen.xend import uuid
from OVSConfig import get_config


def run_cmd(cmd, success_return_code=(0,)):
    if not isinstance(cmd, list):
        raise Exception("Only accepts list!")
    cmd = [str(x) for x in cmd]
    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE, close_fds=True)
    (stdoutdata, stderrdata) = proc.communicate()
    if proc.returncode not in success_return_code:
        raise RuntimeError('Command %s failed (%s): stderr: %s, stdout: %s' %
                (cmd, proc.returncode, stderrdata, stdoutdata))
    return stdoutdata


def old_run_cmd(cmd, success_return_code=(0,), background=False):
    """Warning: incorporate unsanitized input from an untrusted source makes a
    program vulnerable to shell injection, a serious security flaw which can
    result in arbitrary command execution.
    """
    if not isinstance(cmd, basestring):
        raise Exception("Only accepts str!")
    if background:
        devnull = open(os.devnull, "rw")
        proc = subprocess.Popen(cmd, stdin=devnull, stdout=devnull,
                                stderr=devnull, shell=True, close_fds=True)
        proc.wait()
    else:
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                   stderr=subprocess.PIPE, shell=True, close_fds=True)
        (stdoutdata, stderrdata) = proc.communicate()
        if proc.returncode not in success_return_code:
            raise RuntimeError('Command %s failed (%s): stderr: %s, stdout: %s' %
                    (cmd, proc.returncode, stderrdata, stdoutdata))
        return stdoutdata


def run_cmd_bg(cmd):
    old_run_cmd(cmd, background=True)


def retry_cmd(cmd, interval=5, count=3):
    # retry if failed
    for i in range(1, count):
        try:
            output = run_cmd(cmd)
            return output
        except:
            time.sleep(interval)
    # retry for the last time
    return run_cmd(cmd)


def well_behavior(decorator):
    """This decorator can be used to turn simple functions
    into well-behaved decorators, so long as the decorators
    are fairly simple. If a decorator expects a function and
    returns a function (no descriptors), and if it doesn't
    modify function attributes or docstring, then it is
    eligible to use this. Simply apply @well_behavior to
    your decorator and it will automatically preserve the
    docstring and function attributes of functions to which
    it is applied."""
    def new_decorator(f):
        g = decorator(f)
        g.__name__ = f.__name__
        g.__doc__ = f.__doc__
        g.__dict__.update(f.__dict__)
        return g
    # Now a few lines needed to make well_behavior itself
    # be a well-behaved decorator.
    new_decorator.__name__ = decorator.__name__
    new_decorator.__doc__ = decorator.__doc__
    new_decorator.__dict__.update(decorator.__dict__)
    return new_decorator


TRACE_EXCEPTION = [
    'dispatch',
    'ovs_async_proc',
    'get_last_boot_time',
    'update_agent_password',
    'send_to_guest',
    'list_vms',
    'remote_power_on',
    'remote_power_off',
]

SCRUB_LIST = ['password', 'passwd']


def scrub_dict(d):
    for key in d:
        for word in SCRUB_LIST:
            if word in key:
                d[key] = '******'


def scrub_and_log(name, *args):
    temp_args = copy.deepcopy(args)
    for arg in temp_args:
        if type(arg) == dict:
            scrub_dict(arg)
    logging.debug("%s: %s" % (name, temp_args))


@well_behavior
def trace(func):
    def wrapper(*args):
        scrub_and_log(func.__name__, *args)
        result = func(*args)
        logging.debug("%s: call completed.", func.__name__)
        return result
    return wrapper


@well_behavior
def catch_error(func):
    def wrapper(*args):
        try:
            return func(*args)
        except Exception, e:
            logging.exception("catch_error: %s" % e)
            raise
    return wrapper


def expose(api):
    @well_behavior
    def wrapper(func):
        def f(*args):
            new_func = catch_error(func)
            if api not in TRACE_EXCEPTION:
                new_func = trace(new_func)
            return new_func(*args)
        f.__name__ = api
        return f
    return wrapper


def get_uuid():
    uuid.uuidFactory = uuid.getUuidUuidgen
    return uuid.createString()


def host_alive(ip):
    status = os.popen("ping -q -c 2 %s" % ip).close()
    return not status


def get_agt_url(hostname):
    config = get_config()
    if config.get("security", "ssl").lower().startswith("enable"):
        protocol = "https"
    else:
        protocol = "http"
    return "%s://%s:%s" % (protocol, hostname, config.get("agent", "port"))
