# Copyright (C) 2006-2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import string
from os.path import join, isdir, isfile

from OVSConstants import LOCAL_SERVER_DB, CLUSTER_CONF, CLUSTER_CONFIGFS, CLUSTER_CONF
from OVSAgentDB import read_item
from OVSDiscoverCommon import SimpleDiscoveryXML, DiscoveryLock
from OVSServerPool import get_ocfs2_cluster_name


def parse_ocfs2_cluster_conf(cfg_file=CLUSTER_CONF):
    """
    @return: {"nodes":[<dict of node>], "cluster": <dict of cluster>, "heartbeat": <dict of heartbeat>}
    """
    conf_dict = { "nodes": [], "cluster": {}, "heartbeat": {} }

    f = open(cfg_file)
    try:
        for line in f:
            line = line.split("#")[0].strip()
            if not line:
                continue
            elif line.startswith("node:"):
                d = {}
                conf_dict["nodes"].append(d)
            elif line.startswith("cluster:"):
                d = {}
                conf_dict["cluster"] = d
            elif line.startswith("heartbeat:"):
                d = {}
                conf_dict["heartbeat"] = d
            else:
                parts = map(string.strip, line.split("="))
                if len(parts) == 2 and parts[0] and parts[1]:
                    if type(d) == type({}):
                        d[parts[0]] = parts[1]
                    else:
                        d.append(parts[1])
    finally:
        f.close()

    return conf_dict


def discover_stored_cluster_info(doc, registered_hostname):

    node_items = [ ("number", "Number"),
                   ("ip_port", "IP_Port"),
                   ("ip_address", "IP_Address"),
                   ("name", "Name"),
                   ("cluster", "Cluster_Name"), ]

    cluster_items = [ ("name", "Name"),
                      ("node_count", "Node_Count"),
                      ("heartbeat_mode", "Heartbeat_Mode"), ]

    heartbeat_items = [ ("region", "Region"),
                        ("cluster", "Cluster"), ]

    # Returns a dict containing a list of node dicts and a cluster dict
    parsed_config = parse_ocfs2_cluster_conf()
    doc.push("Stored")

    cluster_values = parsed_config["cluster"]
    if not cluster_values:
        raise Exception("cluster.conf 'cluster:' section is invalid or missing")

    for (file_kw, element) in cluster_items:
        if file_kw in cluster_values:
            doc.add(element, cluster_values[file_kw])
        else:
            raise Exception("cluster.conf missing '%s' line" % file_kw)

    doc.push("Heartbeat")

    heartbeat_values = parsed_config["heartbeat"]
    if not heartbeat_values:
        raise Exception("cluster.conf 'heartbeat:' section is invalid or missing")

    for (file_kw, element) in heartbeat_items:
        if file_kw in heartbeat_values:
            doc.add(element, heartbeat_values[file_kw])
        else:
            raise Exception("cluster.conf missing '%s' line" % file_kw)

    doc.pop() # Heartbeat

    node_list = parsed_config["nodes"]
    if not node_list:
        raise Exception("cluster.conf 'node:' section is invalid or missing")

    if registered_hostname:
        for node in node_list:
            if "name" in node and node["name"] == registered_hostname:
                doc.add("Local_Node_Number", node["number"])

    doc.push("Nodes")

    for node in node_list:
        doc.push("Node")
        for (file_kw, element) in node_items:
            if file_kw in node:
                doc.add(element, node[file_kw])
            else:
                raise Exception("cluster.conf missing '%s' line for node %s" %
                                (file_kw, node))
        doc.pop() # Node

    doc.pop() # Nodes

    doc.pop() # Stored


def get_active_cluster_info(cluster_name, registered_hostname):
    mybase = join(CLUSTER_CONFIGFS, cluster_name, "node", registered_hostname)
    if not isdir(mybase):
        return (None, None, None)

    fp = open(join(mybase, "local"), "r")
    if fp.readline().strip() != "1":
        raise Exception("Cluster sysfs entry for this node not marked local")
    fp.close()

    fp = open(join(mybase, "num"), "r")
    num = fp.readline().strip()
    if not num.isdigit():
        raise Exception("Cluster sysfs node number not a number")
    fp.close()

    fp = open(join(mybase, "ipv4_address"), "r")
    ip_addr = fp.readline().strip()
    if not ip_addr:
        raise Exception("Cluster sysfs IP address missing")
    fp.close()

    fp = open(join(mybase, "ipv4_port"), "r")
    ip_port = fp.readline().strip()
    if not ip_port:
        raise Exception("Cluster sysfs IP port missing")
    fp.close()

    return (num, ip_addr, ip_port)


def discover_cluster():
    """
    Discover information about the cluster configured for this server.

    Returns both the stored information from /etc/ocfs2/cluster.conf, as
    well as the server's info if it is actively participating in the cluster.
    Returns an empty document if the server is not configured in a cluster.

    @param      None

    @return     Returns the discovery data as an XML document
                Raises an exception on failure.
    """

    lock = DiscoveryLock("discover_cluster")
    if not lock.acquire():
        raise Exception("discover_cluster already in progress")

    try:
        doc = SimpleDiscoveryXML("Discover_Cluster_Result")

        if isfile(CLUSTER_CONF):

            doc.push("Cluster")

            hostname = read_item(LOCAL_SERVER_DB, "registered_hostname")

            try:
                discover_stored_cluster_info(doc, hostname)
            except Exception, e:
                raise Exception("Unable to discover "
                                "stored cluster info: %s" % e)

            cluster_name = get_ocfs2_cluster_name()

            if hostname and cluster_name:
                try:
                    (num, ip_addr, ip_port) = \
                        get_active_cluster_info(cluster_name, hostname)
                except Exception, e:
                    raise Exception("Unable to discover "
                                    "active cluster info: %s" % e)
                if num != None:
                    doc.push("Active")
                    doc.add("Local_Node_Number", num)
                    doc.add("Local_IP_Address", ip_addr)
                    doc.add("Local_IP_Port", ip_port)
                    doc.pop() # Active

            doc.pop() # Cluster

        try:
            result = doc.finish()
        except Exception, e:
            raise Exception("Unable to format discover_cluster XML: %s" % e)

    finally:
        lock.release()

    return result
