# Copyright (C) 2006-2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import logging
import fcntl
import os
from xml.dom.minidom import Document
from xml.dom.ext import PrettyPrint
from StringIO import StringIO

import OVSDiscoverExtension
from OVSConfig import get_config

XML_INDENT = 2    # Indent amount for XML documents
STORAGE_DISCOVER_PROG = "storage_discover" # External storage discovery utility
SCSI_INFO_DEBUG_LVL = 0 # Default debug level for above util
DISCOVER_LOCK_FILE_TEMPLATE = "/var/run/ovs-agent/%s.lock"

class SimpleDiscoveryXML:
    """
    A simple XML document of the form needed for discovery results.

    Provides for pushing (nesting) elements and adding elements with a
    single value.  The document can be formatted as a pretty-printed
    string.
    All errors are reported by raising exceptions.
    """
    def __init__(self, toplevel):
        # Create a XML document starting with the specified element
        if not toplevel:
            raise Exception("A top level element name is required")
        self.document = Document()
        self.elementstack = []
        top = self.document.createElement(toplevel)
        self.elementstack.append(top)
        self.document.appendChild(top)

    def add(self, element, value=None):
        elem = self.document.createElement(element)
        top = self.elementstack[-1]
        top.appendChild(elem)
        if value != None:
            elem.appendChild(self.document.createTextNode(value))

    def push(self, element, attribute=None, attr_value=None):
        newtop = self.document.createElement(element)
        oldtop = self.elementstack[-1]
        self.elementstack.append(newtop)
        oldtop.appendChild(newtop)
        if attribute:
            newtop.setAttribute(attribute, attr_value)

    def pop(self):
        if len(self.elementstack) == 0:
            raise Exception("Popped empty element stack in SimpleDiscoveryXML")
        self.elementstack.pop()

    def finish(self):
        if len(self.elementstack) > 1:
            raise Exception("Unpopped element in SimpleDiscoveryXML")

        # Generate the formatted XML document
        stream = StringIO()
        PrettyPrint(self.document, stream)
        result = stream.getvalue()
        return result

class DiscoveryLock:
    """
    A simple lock (based on flock) to serialize discovery operations.
    """
    def __init__(self, lockname):
        self.lockname = lockname
        self.filename = DISCOVER_LOCK_FILE_TEMPLATE % lockname
        self.lockfile = None

    def acquire(self):
        """
        @return True if the lock was acquired or False if a conflict
        """
        try:
            self.lockfile = open(self.filename, "a")
            fcntl.flock(self.lockfile, fcntl.LOCK_EX | fcntl.LOCK_NB)
            self.lockfile.write("%d" % os.getpid())
            self.lockfile.flush()
            return True
        except:
            self.lockfile.close()
            self.lockfile = None
            return False

    def release(self):
        try:
            fcntl.flock(self.lockfile, fcntl.LOCK_UN)
            self.lockfile.close()
            os.remove(self.filename)
        except:
            pass

    def __del__(self):
        self.release()


## Deprecated - only used by obsolete(?) Discover.status()
def format_extn_response(status, response, logmsg):
    """
    Construct an XMLRPC response from the standard tuple returned by all
    routines in the C extension.  Log any logmsg returned.

    @param status       0 on success, non-zero on failure
    @param response     the result from the routine (data or error message)
    @param logmsg       optional message to be written to the log on failure
    """

##    print "(%d) %s\n" % (status, logmsg)

    if status != 0:
        word = "failed:"
        if logmsg != None:
            logging.error(logmsg)
    else:
        word = "success:"
    return word + response

def get_server_unique_id():
    """
    Obtain a globally unique identifier for this server.

    This invokes legacy C code from the VI agent and returns a 16 byte
    unique ID in the form of a string of hex digits:
        xx:xx:xx:xx:xx:xx:xx:xx:xx:xx:xx:xx:xx:xx:xx:xx

    @return    The UUID string or None on error
    """
    config = get_config()
    try:
        fakeuuid = config.get("server","fakeuuid")
    except:
        fakeuuid = ""

    if (fakeuuid != ""):
        response = fakeuuid
        return response
    else:
        (status, response, logmsg) = OVSDiscoverExtension.unique_id(None)
        if (status == 0):
            return response
        else:
            logging.error(logmsg)
            return None
