# Copyright (C) 2006-2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import subprocess

from OVSDiscoverCommon import (DiscoveryLock, STORAGE_DISCOVER_PROG,
                               SCSI_INFO_DEBUG_LVL)

import OVSDiscoverExtension

DISCOVERY_MARKER = "##STORAGE_DISCOVERY##\n"

def scsi_debug_arg():
    if SCSI_INFO_DEBUG_LVL > 0:
        return "--debugLvl %d" % SCSI_INFO_DEBUG_LVL
    else:
        return ""

def do_scsi_no_detail(indent):

    # The output here will get inserted into already formatted discover output
    args = [STORAGE_DISCOVER_PROG, scsi_debug_arg(),
            "--indent", "%d" % indent,
            "--infoDumpType", "io_host_XML"]

    proc = subprocess.Popen(args, stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE, close_fds=True)
    (stdout_string, stderr_string) = proc.communicate()
    return (proc.returncode, stdout_string, stderr_string)

def discover_hardware():
    """
    Discover information about the hardware present on this server.
    This includes CPU, I/O Adapters, Memory, and BIOS info.
    No information about storage devices is returned.

    Discovery of storage adapters is handled from python by forking and running
    the storage discovery utility.  Other discovery is handled by
    invoking a python extension that calls legacy C code from the VI agent.

    @param None

    @return    Returns the discovery data as an XML document.
               Raises an exceptionon failure.
    """

    lock = DiscoveryLock("discover_hardware")
    if not lock.acquire():
        raise Exception("discover_hardware already in progress")

    # Get the non-storage hardware, complete with headers
    (status, response, logmsg) = OVSDiscoverExtension.discover_hardware("")
    if status != 0:
        lock.release()
        raise Exception(logmsg)

    # Find the marker for where to insert storage data
    idx = response.find(DISCOVERY_MARKER)
    marker_begin = response.rfind("\n", 0, idx)

    if idx < 0 or marker_begin < 0 or marker_begin >= idx:
        lock.release()
        raise Exception("Internal Error: No storage discovery marker")

    indent = idx - marker_begin -1
    whole_marker = " " * indent + DISCOVERY_MARKER

    # Get the storage data
    (scsi_status, scsi_response, scsi_logmsg) = do_scsi_no_detail(indent)
    if scsi_status != 0:
        lock.release()
        raise Exception(scsi_logmsg)

    result = response.replace(whole_marker, scsi_response)
    lock.release()
    return result
