# Copyright (C) 2006-2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import os
import codecs
import xml.dom.minidom
import OVSDiscoverExtension
from OVSDiscoverCommon import DiscoveryLock
from xml.dom.ext import PrettyPrint
from StringIO import StringIO
from utils import netutils


def get_meta(device):
    filename = '/etc/sysconfig/network-scripts/meta-%s' % device
    if os.path.exists(filename):
        metafile = codecs.open(filename, encoding='utf-8')
        meta = metafile.readlines()
        metafile.close()
        for line in meta:
            if line.startswith('METADATA='):
                return line[len('METADATA='):]


def discover_network():
    """
    Discover information about the current network configuration.
    This includes the state of physical NICs, bonds, and bridges.  Also return
    information stored for this server that is needed to configure
    the network when the OVM Manager is not available.

    Discovery of the current network configuration is handled by invoking
    a python extension that calls legacy C code from the VI agent.

    @param None

    @return    Returns the discovery data as an XML document.
               Raises an exception on failure.
    """
    lock = DiscoveryLock("discover_network")
    if not lock.acquire():
        raise Exception("discover_network already in progress")
    try:
        (status, network_xml, logmsg) = OVSDiscoverExtension.discover_network(0)
        if status != 0:
            raise Exception(logmsg)
        result = '<Discover_Network_Result><Network><Active>%s</Active></Network></Discover_Network_Result>' % network_xml
        doc = xml.dom.minidom.parseString(result)
        for node in doc.getElementsByTagName("Device"):
            device = node.getAttribute("Name")
            config = netutils.getConfig(device)
            if "BOOTPROTO" in config and config["BOOTPROTO"] in ["dhcp", "none", "static"]:
                proto = doc.createElement("BootProto")
                proto_text = doc.createTextNode(config["BOOTPROTO"])
                proto.appendChild(proto_text)
                node.appendChild(proto)
            meta = get_meta(device)
            if meta:
                metadata = doc.createElement("MetaData")
                metadata_text = doc.createTextNode(meta)
                metadata.appendChild(metadata_text)
                node.appendChild(metadata)
        stream = StringIO()
        PrettyPrint(doc, stream)
        pretty_result = stream.getvalue()
        stream.close()
        return pretty_result
    finally:
        lock.release()


def discover_saved_network():
    """
    Discover saved network configurations under /etc/sysconfig/network-scripts/ifcfg-*.

    @return    Returns the discovery data as an XML document.
               Raises an exception on failure.
    """
    lock = DiscoveryLock("discover_saved_network")
    if not lock.acquire():
        raise Exception("discover_network already in progress")
    try:
        doc = xml.dom.minidom.Document()
        top = doc.createElement("Discover_Network_Result")
        doc.appendChild(top)
        network_first = doc.createElement("Network")
        top.appendChild(network_first)
        stored = doc.createElement("Saved")
        network_first.appendChild(stored)
        network_second = doc.createElement("Network")
        stored.appendChild(network_second)
        bonding = doc.createElement("Bonding")
        stored.appendChild(bonding)
        bridges = doc.createElement("Bridges")
        stored.appendChild(bridges)

        for device in netutils.getConfiguredDevices():
            config = netutils.getConfig(device)
            if "DEVICE" not in config or config["DEVICE"] != device or device == "lo":
                continue
            node = doc.createElement("Device")
            node.setAttribute("Name", device)
            if config.get("TYPE") == "Bridge":
                interfaces = doc.createElement("Interfaces")
                for port in netutils.getBridgePortsFromConfig(device):
                    phy_if = doc.createElement("PhyInterface")
                    phy_if_text = doc.createTextNode(port)
                    phy_if.appendChild(phy_if_text)
                    interfaces.appendChild(phy_if)
                node.appendChild(interfaces)
                bridges.appendChild(node)
            elif device in netutils.getBondMastersFromConfig():
                interfaces = doc.createElement("Interfaces")
                for slave in netutils.getBondSlavesFromConfig(device):
                    slave_if = doc.createElement("Slave_Interface")
                    slave_if.setAttribute("Name", slave)
                    node.appendChild(slave_if)
                bonding.appendChild(node)
            else:
                network_second.appendChild(node)
            if "HWADDR" in config:
                mac = doc.createElement("MAC")
                mac_text = doc.createTextNode(config["HWADDR"])
                mac.appendChild(mac_text)
                node.appendChild(mac)
            if "BOOTPROTO" in config and config["BOOTPROTO"] in ["dhcp", "none", "static"]:
                proto = doc.createElement("BootProto")
                proto_text = doc.createTextNode(config["BOOTPROTO"])
                proto.appendChild(proto_text)
                node.appendChild(proto)
            if "IPADDR" in config:
                family = doc.createElement("Family")
                family.setAttribute("Type", "AF_INET")
                ip = doc.createElement("Address")
                ip_text = doc.createTextNode(config["IPADDR"])
                ip.appendChild(ip_text)
                family.appendChild(ip)
                if "NETMASK" in config:
                    netmask = doc.createElement("Netmask")
                    netmask_text = doc.createTextNode(config["NETMASK"])
                    netmask.appendChild(netmask_text)
                    family.appendChild(netmask)
                if "BROADCAST" in config:
                    broadcast = doc.createElement("Broadcast")
                    broadcast_text = doc.createTextNode(config["BROADCAST"])
                    broadcast.appendChild(broadcast_text)
                    family.appendChild(broadcast)
                node.appendChild(family)
            meta = get_meta(device)
            if meta:
                metadata = doc.createElement("MetaData")
                metadata_text = doc.createTextNode(meta)
                metadata.appendChild(metadata_text)
                node.appendChild(metadata)

        stream = StringIO()
        PrettyPrint(doc, stream)
        result = stream.getvalue()
        stream.close()
        lock.release()
        return result
    finally:
        lock.release()
