# Copyright (C) 2006-2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import subprocess
from os.path import exists
from OVSDiscoverCommon import (XML_INDENT, STORAGE_DISCOVER_PROG,
                               SCSI_INFO_DEBUG_LVL, DiscoveryLock)

def XML_tag_with_value(element, indent, value):
    template = "%-*s<%s>%s</%s>\n"
    if element == "" or indent < 0:
        return "? invalid arguments"

    return template % (indent, "", element, value, element)


def XML_tag(element, indent, start):
    template = "%-*s<%s%s>\n"
    if element == "" or indent < 0:
        return "? invalid arguments"

    if start:
        close = ""
    else:
        close = "/"
    return template % (indent, "", close, element)


def XML_header(indent):
    header = '<?xml version="1.0" encoding="UTF-8"?>\n'
    indent += XML_INDENT

    header = header + XML_tag("Discover_Physical_LUNS_Result", indent, True)
    indent += XML_INDENT
    return (indent, header)

def XML_trailer(indent):
    indent -= XML_INDENT
    trailer = XML_tag("Discover_Physical_LUNS_Result", indent, False)
    return (indent, trailer)

def scsi_debug_arg():
    if SCSI_INFO_DEBUG_LVL > 0:
        return "--debugLvl %d" % SCSI_INFO_DEBUG_LVL
    else:
        return ""

def discover_physical_luns(args):
    """
    Discover information about one or more LUNs visible on this server.
    If an empty argument string is specified, return information about all LUNs

    @param      Space-separated list of SCSI IDs to report on

    @return    If successful, returns the discovery data in XML
               Raises an exception on failure.
    """

    lock = DiscoveryLock("discover_physical_luns")
    if not lock.acquire():
        raise Exception("discover_physical_luns already in progress")

    disk_list = args

    indent = 0
    (indent, header) = XML_header(indent)

    prog_args = [STORAGE_DISCOVER_PROG, scsi_debug_arg(),
                 "--indent", "%d" % indent,
                 "--infoDumpType", "disk_XML" ]

    if disk_list != "":
        # List is sent via stdin
        disk_list_string = disk_list.replace(" ", "\n") + "\n"
        prog_args.extend(["--diskList", "<<STDIN>>"])
    else:
        disk_list_string = ""

    proc = subprocess.Popen(prog_args, stdin=subprocess.PIPE,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE, close_fds=True)

    (stdout_string, stderr_string) = proc.communicate(disk_list_string)

    exclusions = ''

    exclusions_list = get_lun_exclusions()
    if exclusions_list:
        exclusions = XML_tag("Exclusions", indent, True)
        for entry in exclusions_list:
            exclusions += XML_tag_with_value("Disk", indent + XML_INDENT, entry)
        exclusions += XML_tag("Exclusions", indent, False)

    
    (indent, trailer) = XML_trailer(indent)

    if proc.returncode != 0 or stderr_string:
        lock.release()
        raise Exception("discover_physical_luns: %s" % stderr_string)

    result = header + stdout_string + exclusions + trailer

    lock.release()
    return result

def get_lun_exclusions():

    luns = []

    if exists("/etc/blacklisted.wwids.ignored"):
        f = open("/etc/blacklisted.wwids.ignored", 'r')
        try:
            lines = f.readlines()
        finally:
            f.close()

        for line in lines:
            line = line.replace('\"',"")
            line = line.replace("wwid", "").strip()
            if not line:
                continue
            luns.append(line)

    return luns

