# Copyright (C) 2006-2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import logging
from os.path import join, exists

from OVSConstants import POOLFS_META, POOLFS_LABEL, LUN_PREFIX
from OVSCommons import run_cmd
from OVSPoolFileSystem import parse_metafile
from OVSDiscoverCommon import SimpleDiscoveryXML, DiscoveryLock
from OVSOCFS2 import ocfs2_file_lookup, ocfs2_file_read


def add_nfs_poolfs_info(doc):
    """
    Add XML for NFS based pool filesystem.
    """
    output = run_cmd(["cat", "/proc/mounts"])
    lines = output.splitlines()

    for line in lines:
        parts = line.split()
        if parts[1].startswith("/nfsmnt/") and parts[2] == "nfs":
            poolfs_meta = join(parts[1], POOLFS_META)
            if exists(poolfs_meta):
                f = file(poolfs_meta)
                try:
                    contents = f.read()
                finally:
                    f.close()
                d = parse_metafile(contents)

                doc.push("Pool_Filesystem")
                doc.add("Pool_Filesystem_Type", "nfs")
                doc.add("Pool_Filesystem_Target", parts[0])
                doc.add("Pool_Filesystem_Uuid", d.get("OVS_POOLFS_UUID"))
                doc.add("Pool_Filesystem_Nfsbase_Uuid",
                        d.get("OVS_POOLFS_NFSBASE_UUID"))
                doc.add("Pool_Filesystem_Manager_Uuid",
                        d.get("OVS_POOLFS_MGR_UUID"))
                doc.add("Pool_Filesystem_Version",
                        d.get("OVS_POOLFS_VERSION"))
                doc.add("Pool_Filesystem_Pool_Uuid",
                        d.get("OVS_POOLFS_POOL_UUID"))
                doc.pop()


def add_lun_poolfs_info(doc):
    """
    Add XML for LUN based pool filesystem.
    """
    output = run_cmd(["mounted.ocfs2", "-d"])
    lines = output.splitlines()

    for line in lines:
        parts = line.split()
        if parts[0].startswith(LUN_PREFIX) and parts[-1] == POOLFS_LABEL:
            try:
                if ocfs2_file_lookup(parts[0], POOLFS_META) != 0:
                    continue
                contents = ocfs2_file_read(parts[0], POOLFS_META)
                d = parse_metafile(contents)

                doc.push("Pool_Filesystem")
                doc.add("Pool_Filesystem_Type", "lun")
                doc.add("Pool_Filesystem_Target", parts[0])
                doc.add("Pool_Filesystem_Uuid", d.get("OVS_POOLFS_UUID"))
                doc.add("Pool_Filesystem_Lun_Uuid",
                        d.get("OVS_POOLFS_LUN_UUID"))
                doc.add("Pool_Filesystem_Manager_Uuid",
                        d.get("OVS_POOLFS_MGR_UUID"))
                doc.add("Pool_Filesystem_Version",
                        d.get("OVS_POOLFS_VERSION"))
                doc.add("Pool_Filesystem_Pool_Uuid",
                        d.get("OVS_POOLFS_POOL_UUID"))
                doc.pop()
            except Exception, e:
                logging.warn('Cannot get poolfs info: %s', e)


def discover_pool_filesystem():
    """
    Discover information about the pool filesystem.
    Returns an empty document if no pool filesystem found.

    @param      None
    @return     Discovery data as an XML document.
                Raises an exception on failure.
    """

    lock = DiscoveryLock("discover_pool_filesystem")
    if not lock.acquire():
        raise Exception("discover_pool_filesystem already in progress")

    try:
        doc = SimpleDiscoveryXML("Discover_Pool_Filesystem_Result")
        add_nfs_poolfs_info(doc)
        add_lun_poolfs_info(doc)
        try:
            result = doc.finish()
        except Exception, e:
            raise Exception("Unable to format "
                            "discover_pool_filesystem XML: %s" % e)
    finally:
        lock.release()

    return result
