# Copyright (C) 2006-2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import socket

import host
import OVSServerPool
from OVSConstants import LOCAL_SERVER_DB, MONITOR_SERVER_INTERVAL
from OVSAgentDB import dump_db
from OVSCommons import run_cmd
from OVSDiscoverCommon import (SimpleDiscoveryXML, DiscoveryLock,
        get_server_unique_id)


def get_rootfs_id():
    response = run_cmd(["/lib/udev/vol_id", "--uuid", "/dev/root"])
    for line in response.splitlines():
        if line.startswith("ID_FS_UUID="):
            return line.split("=")[1]
    raise Exception("No root filesystem UUID found")


def discover_xen(doc):
    # Interesting items from "xm info"
    kernel_items = [ ("release", "Host_Kernel_Release"),
                     ("version", "Host_Kernel_Version"), ]

    version_items = [ ("xen_major", "Major"),
                      ("xen_minor", "Minor"),
                      ("xen_extra", "Extra"), ]

    xen_items = [ ("xen_caps", "Capabilities"),
                  ("xen_changeset", "ChangeSet"), ]

    compiler_items = [ ("cc_compiler", "Compiler"),
                       ("cc_compile_by", "By"),
                       ("cc_compile_domain", "Domain"),
                       ("cc_compile_date", "Date"), ]

    # Build a dictionary of attributes and values returned by "xm info"
    response = run_cmd(["xm", "info"])

    xminfo = {}
    for line in response.splitlines():
        (attr, val) = line.split(':', 1)
        xminfo[attr.strip()] = val.strip()

    for (xminfo_name, discovery_name) in kernel_items:
        doc.add(discovery_name, xminfo[xminfo_name])

    doc.push("VMM")
    doc.push("Version")
    for (xminfo_name, discovery_name) in version_items:
        doc.add(discovery_name, xminfo[xminfo_name])
    doc.pop()

    doc.push("Compile_Information")
    for (xminfo_name, discovery_name) in compiler_items:
        doc.add(discovery_name, xminfo[xminfo_name])
    doc.pop()

    for (xminfo_name, discovery_name) in xen_items:
        doc.add(discovery_name, xminfo[xminfo_name])

    doc.pop()  # VMM


def discover_server():
    """
    Discover the identity, version and other properties of the server.

    @param      None

    @return     Discovery data as an XML document.
                Raises an exception on failure.
    """

    lock = DiscoveryLock("discover_server")
    if not lock.acquire():
        raise Exception("discover_server already in progress")

    try:
        doc = SimpleDiscoveryXML("Discover_Server_Result")
        doc.push("Server")

        response = get_server_unique_id()
        if not response:
            raise Exception("Unable to obtain server unique id")

        doc.add("Unique_Id", response)

        try:
            response = str(OVSServerPool.get_last_boot_time()['last_boot_time'])
        except Exception, e:
            raise Exception("Unable to obtain last boot time: %s" % e)

        doc.add("Boot_Time", response)

        (timezone, utc) = host.get_timezone()
        doc.push("Date_Time")
        doc.add("Time_Zone", str(timezone))
        doc.add("UTC", str(utc))
        doc.pop()

        (ntp_servers, local_time_source, is_ntp_running) = host.get_ntp()
        doc.push("NTP")
        for ntp_server in ntp_servers:
            doc.add("Server", ntp_server)
        doc.add("Local_Time_Source", str(local_time_source))
        doc.add("Is_NTP_Running", str(is_ntp_running))
        doc.pop()

        try:
            rootfs_id = get_rootfs_id()
        except Exception, e:
            raise Exception("Unable to discover rootfs id: %s" % e)

        doc.add("Root_Filesystem_Id", rootfs_id)

        pkgs = host.list_package('ovs-agent')
        if not pkgs:
            raise Exception("unable to find package ovs-agent.")
        agent_version = pkgs[0]['version']
        agent_release = pkgs[0]['release']
        doc.add("Agent_Version", '-'.join([agent_version, agent_release]))
        # RPM and agent versions are currently the same
        doc.add("RPM_Version", '-'.join([agent_version, agent_release]))

        doc.add('OVM_Version', host.get_ovs_version()['release'])

        try:
            discover_xen(doc)
        except Exception, e:
            raise Exception("Unable to discover xen information: %s" % e)

        uuid = OVSServerPool.get_server_pool_uuid()
        if not uuid:
            uuid = "none"
        doc.add("Pool_Unique_Id", uuid)

        uuid = OVSServerPool.get_owning_manager_uuid()
        if not uuid:
            uuid = "none"
        doc.add("Manager_Unique_Id", uuid)
        doc.add("Hostname", socket.gethostname())

        d = dump_db(LOCAL_SERVER_DB)
        registered_hostname = d.get("registered_hostname")
        registered_ip = d.get("registered_ip")
        node_number = d.get("node_number")
        server_roles = d.get("roles")
        is_master = d.get("is_master")
        virtual_ip = d.get("pool_virtual_ip")
        stat_interval = d.get("stat_interval")

        if registered_hostname:
            doc.add("Registered_Hostname", registered_hostname)

        if registered_ip:
            doc.add("Registered_IP", registered_ip)

        if node_number != None:
            doc.add("Node_Number", str(node_number))

        if not server_roles:
            server_roles = "unassociated"
        else:
            server_roles = ",".join(list(server_roles))

        doc.add("Server_Roles", server_roles)

        if is_master == True:
            master_str = "true"
        else:
            master_str = "false"
        doc.add("Is_Current_Master", master_str)

        if virtual_ip:
            doc.add("Master_Virtual_Ip", virtual_ip)

        doc.add("Manager_Core_API_Version", host.get_core_api_version() or '0.0.0.0')

        doc.add("Membership_State", OVSServerPool.get_membership_state())
        doc.add("Cluster_State", OVSServerPool.get_cluster_state())

        # Statistic
        stat_interval = stat_interval or MONITOR_SERVER_INTERVAL
        doc.push("Statistic")
        doc.add("Interval", str(stat_interval))
        doc.pop()

        # Server
        doc.pop()

        try:
            result = doc.finish()
        except Exception, e:
            raise Exception("Unable to format "
                            "discover_server XML: %s" % str(e))
    finally:
        lock.release()

    return result
