# Copyright (C) 2006-2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

from OVSConstants import *
from OVSAgentDB import dump_db, get_cluster_db_home
from OVSDiscoverCommon import SimpleDiscoveryXML, DiscoveryLock
from OVSServerPool import get_server_pool_uuid, is_clustered


def add_clustered_info(doc):
    """
    Add XML for server pool members and virtual IP if clustered
    """
    d = dump_db("server_pool", get_cluster_db_home())
    pool_alias = d.get("pool_alias")
    master_vip = d.get("pool_virtual_ip")
    member_list = d.get("pool_member_ip_list")

    if pool_alias:
        doc.add("Pool_Alias", pool_alias)

    if master_vip:
        doc.add("Master_Virtual_Ip", master_vip)

    doc.push("Member_List")

    for ip in member_list:
        doc.push("Member")
        doc.add("Registered_IP", ip)
        doc.pop() # Member

    doc.pop() # Member_List


def add_pooled_info(doc):
    """
    Add XML for server pool members that we know about when not yet clustered
    """
    d = dump_db(LOCAL_SERVER_DB)
    pool_alias = d.get("pool_alias")
    virtual_ip = d.get("pool_virtual_ip")
    member_list = d.get("pool_member_ip_list")

    if pool_alias:
        doc.add("Pool_Alias", pool_alias)

    if virtual_ip:
        doc.add("Master_Virtual_Ip", virtual_ip)

    doc.push("Member_List")

    if member_list:
        for srv in member_list:
            doc.push("Member")
            doc.add("Registered_IP", srv)
            doc.pop() # Member

    doc.pop() # Member_List


def add_poolfs_info(doc):
    """Add XML for the pool filesystem if available."""
    d = dump_db(LOCAL_SERVER_DB)
    poolfs_type = d.get("poolfs_type")
    poolfs_target = d.get("poolfs_target")
    poolfs_uuid = d.get("poolfs_uuid")
    poolfs_nfsbase_uuid = d.get("poolfs_nfsbase_uuid")

    if poolfs_type:
        doc.push("Pool_Filesystem")
        doc.add("Pool_Filesystem_Type", poolfs_type)
        doc.add("Pool_Filesystem_Target", poolfs_target)
        doc.add("Pool_Filesystem_Uuid", poolfs_uuid)
        doc.add("Pool_Filesystem_Nfsbase_Uuid", poolfs_nfsbase_uuid)
        doc.pop() # Pool_Filesystem


def discover_server_pool():
    """
    Discover information about the server pool that contains this server.

    Returns an empty document if the server is unassociated.

    @param      None

    @return     If successful, returns the discovery data as an XML document
                Raises an exception on failure.
    """

    lock = DiscoveryLock("discover_server_pool")
    if not lock.acquire():
        raise Exception("discover_server_pool already in progress")

    try:
        doc = SimpleDiscoveryXML("Discover_Server_Pool_Result")
        uuid = get_server_pool_uuid()

        if uuid:
            doc.push("Server_Pool")
            doc.add("Unique_Id", uuid)

            add_poolfs_info(doc)

            if is_clustered():
                add_clustered_info(doc)
            else:
                add_pooled_info(doc)

            doc.pop() # Server_Pool

        try:
            result = doc.finish()
        except Exception, e:
            raise Exception("Unable to format "
                            "discover_server_pool XML: %s" % str(e))
    finally:
        lock.release()

    return result
