#!/usr/bin/env python

# Copyright (C) 2006-2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import os
import socket
import logging
from multiprocessing import Pool
from time import sleep

from OVSConstants import *
from OVSLog import init_log
from OVSDaemon import daemonize
from OVSConfig import get_config
from OVSAgentDB import dump_db, read_item, delete_item, get_cluster_db_home
from OVSServerPool import is_clustered
from OVSMasterFailover import is_master, dlm_acquire_lock, dlm_release_lock
from OVSCommons import get_agt_url
from utils.xmlrpclib import ServerProxy
from vm import get_vm_path, read_cfg, format_uuid, is_running


def f(hostname):
    try:
        sp = ServerProxy(get_agt_url(hostname))
        vms = sp.list_vms()
        dom0_uuid = "00000000-0000-0000-0000-000000000000"
        return [ vm["uuid"] for vm in vms if vm["uuid"] != dom0_uuid ]
    except Exception, e:
        logging.error("Failed to get VM list on %s: %s", hostname, e)


def get_live_vms():
    hostnames = dump_db("server_pool_servers", get_cluster_db_home()).keys()
    pool = Pool(processes=len(hostnames))
    results = pool.map(f, hostnames)
    live_vms = []
    for e in results:
        if e:
            live_vms.extend(e)
    return live_vms


def get_monitored_vms():
    return dump_db("monitored_vms", get_cluster_db_home()).keys()


def is_vm_running(repo_id, vm_id):
    try:
        path = get_vm_path(repo_id, vm_id)
        cfg = read_cfg(path)
        # is it running on this server?
        if is_running(cfg["name"]):
            return True
        # is it running somewhere else?
        uuid = format_uuid(cfg["uuid"])
        try:
            dlm_acquire_lock(uuid)
            dlm_release_lock(uuid)
            return False
        except:
            return True
    except Exception, e:
        logging.debug("Error checking if VM is running: %s", e)
        # assume it is not running
        return False


def ha_start_vm(repo_id, vm_id):
    logging.debug("Trying to start VM %s", vm_id)
    hostnames = dump_db("server_pool_servers", get_cluster_db_home()).keys()
    for hostname in hostnames:
        try:
            if is_vm_running(repo_id, vm_id):
                logging.debug("VM %s already started", vm_id)
                return
            sp = ServerProxy(get_agt_url(hostname))
            sp.start_vm(repo_id, vm_id)
            logging.debug("VM %s started on %s", vm_id, hostname)
            return
        except Exception, e:
            logging.error("Failed to start VM %s on %s: %s", vm_id, hostname, e)
    logging.error("Cannot start VM %s: all attempts failed", vm_id)


def apply_ha_policies(uuid_list):
    for uuid in uuid_list:
        cluster_db_home = get_cluster_db_home()
        vm = read_item("monitored_vms", uuid, cluster_db_home)
        if os.path.exists(get_vm_path(vm['repo_id'], vm['vm_id'])):
            ha_start_vm(vm['repo_id'], vm['vm_id'])
        else:
            delete_item("monitored_vms", uuid, cluster_db_home)


def run():
    logging.info("HA SERVER STARTED")
    config = get_config()
    ha_interval = int(config.get("ha", "interval"))
    ha_timeout = int(config.get("ha", "timeout"))
    # set socket default timeout for all connections
    # which prevents xmlrpc from hanging infinitely
    socket.setdefaulttimeout(ha_timeout)
    while True:
        try:
            sleep(ha_interval)
            if is_clustered() and is_master():
                live_vms = get_live_vms()
                monitored_vms = get_monitored_vms()
                diff = set(monitored_vms) - set(live_vms)
                apply_ha_policies(diff)
        except Exception, e:
            logging.exception("Error in HA process: %s", e)
            sleep(SLEEP_ON_ERR)


def main():
    pid_filename = os.path.basename(__file__).split(".")[0]
    daemonize(pid_filename)
    init_log()
    try:
        run()
    except Exception, e:
        logging.exception("Error in HA process: %s", e)


if __name__ == "__main__":
    main()
