# Copyright (C) 2006-2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import os
import time

from OVSCommons import run_cmd, retry_cmd


def shutdown(reboot=False):
    """
    Shutdown or reboot the server.

    @return     None. Raises an exception on failure.
    """
    try:
        # shutdown all running vms
        run_cmd(["xm", "shutdown", "-a"])

        # Ought to be a better way to check for vm termination
        # wait up to 2 minutes for the xm list to empty then
        # force the issue
        timeout = 120
        while timeout > 0:
            stdout = run_cmd(["xm", "list"])
            lines = stdout.splitlines()
            if len(lines) < 3:
                break
            time.sleep(10)
            timeout -= 10
    except:
        pass

    if reboot:
        cmd = ["shutdown", "-r", "now", "reboot requested by OVM"]
    else:
        cmd = ["shutdown", "-P", "-h", "now", "shutdown requested by OVM"]

    run_cmd(cmd)


def sys_shutdown():
    """
    Shutdown the server.

    @return     None. Raises an exception on failure.
    """
    shutdown()


def sys_reboot():
    """
    Reboot the server.

    @return     None. Raises an exception on failure.
    """
    shutdown(reboot=True)


def remote_power_off(controller_type, tgt_host, bmc_username="", bmc_password=""):
    """
    Shutdown using ipmitool. For whatever reasons it's been decided
    that this can also shutdown another host so we need to take a
    hostname/ip as a paramater.

    @param      controller_type (fixed as IPMI)
    @param      tgt_host
    @param      bmc_username
    @param      bmc_password

    @return     Node. Raises an exception on failure.
    """
    if controller_type != "IPMI":
        raise Exception("Controller type not supported: %s" % controller_type)

    cmd = ["ipmitool", "-I", "lan", "-H", tgt_host]
    if bmc_username:
        cmd.extend(["-U", bmc_username])
    if bmc_password:
        cmd.extend(["-P", bmc_password])
    cmd.extend(["power", "off"])

    # It transpires that Dell BMC's are somewhat unreliable
    # (actually it may not just be exclusive to them but the result is
    # that we need to send this command a number of times to make sure it
    # was recieved)
    # leave in a small interval for the card to wake up and process the command
    try:
        retry_cmd(cmd, interval=5, count=5)
    except:
        raise Exception("Remote power off failed: %s, %s" %
                (controller_type, tgt_host))


def sys_wol(tgt_mac, net_device="eth0"):
    """
    Attempt to wake the host with ethernet packet(with optional message)
    This uses ether-wake so the target machine *MUST* have the network
    card in etherwake mode

    This is typically achieved with "ethtool -s ethX wol g" on the target host
    before shutdown. Also be aware that if standby power to the machine is
    lost, then this setting will usually be lost and the machine will not be
    able to be revived with this method. A few motherboards do allow the wol
    settings to be preserved after a powerloss. Basically this is a rather
    hit or miss affair.

    @param      tgt_mac (MAC format: aa:bb:cc:dd:ee:ff)
    @param      net_device (default: eth0)
    """
    # It should just work with one packet but just for good measure
    # we'll send it out a few more times to be sure
    cmd = ["ether-wake", "-i", net_device, tgt_mac]
    retry_cmd(cmd, interval=1, count=5)


def remote_power_on(controller_type, tgt, arg1="", arg2=""):
    """
    Attempt to wakup machine over the lan using the ipmitool utility
    basically this is the same as wol/ether-wake etc.

    controller_type (IPMI, WOL, <future extensions>, ...)
    tgt [IP] or just tgt [MAC addr]
    arg1 = bmc_username [for IPMI] net_device [for WOL]
    arg2 = bmc_password [for IPMI only]

    e.g.,
    4 args   -   IPMI, 1.2.3.4, fred, flintstone
    3 args   -   WOL, 11:22:33:44:55:66, eth1
    """
    if controller_type not in ["IPMI", "WOL"]:
        raise Exception("Controller type not supported: %s" % controller_type)

    # So why not call this directly from the manager side?
    # It turns out the licencing is somehow imcompatable
    # for bundling with the manager.

    if controller_type == "WOL":
        # Not necessary to do this, it's just for readability
        net_device = arg1
        sys_wol(tgt, net_device)
    else:
        # For readability...
        bmc_username = arg1
        bmc_password = arg2
        cmd = ["ipmitool", "-I", "lan", "-H", tgt]
        if bmc_username:
            cmd.extend(["-U", bmc_username])
        if bmc_password:
            cmd.extend(["-P", bmc_password])
        cmd.extend(["power", "on"])

        # It transpires that Dell BMC's are somewhat unreliable
        # (actually it may not just be exclusive to them but the result is
        # that we need to send this command a number of times to make sure it
        # was recieved)
        try:
            retry_cmd(cmd, interval=5, count=5)
        except:
            raise Exception("Remote power on failed: %s, %s" %
                    (controller_type, tgt))
