#!/usr/bin/env python

# Copyright (C) 2006-2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import os
import sys
import time
import errno
import Queue
import signal
import struct
import socket
import logging
import threading
import cPickle
from os.path import join, basename
from OVSDaemon import daemonize
from urlparse import urlparse
from SocketServer import ThreadingUnixStreamServer, StreamRequestHandler
from OVSServerPool import get_manager_core_api_url
from OVSDiscoverCommon import get_server_unique_id
from OVSLog import init_log
from OVSConstants import *


OvmClient = None
Server = None
PIDFILE = None
QUEUE = Queue.Queue(1024 * 1024)
SERVER_OWNED = False
_g_uuid = None
_g_eventmgr = None
_g_server = None
_g_statsmgr = None
unix_server = None


def load_core_api_bindings():
    global OvmClient
    global Server
    try:
        from com.oracle.ovm.mgr.api import OvmClient
        from com.oracle.ovm.mgr.api.physical import Server
        reload(OvmClient)
        reload(Server)
    except ImportError:
        OvmClient = Server = None
        logging.debug('Error loading core api bindings.')


class MsgHandler(StreamRequestHandler):
    def handle(self):
        while True:
            try:
                chunk = self.connection.recv(4)
                if len(chunk) < 4:
                    break
                slen = struct.unpack(">L", chunk)[0]
                chunk = self.connection.recv(slen)
                while len(chunk) < slen:
                    chunk = chunk + self.connection.recv(slen - len(chunk))
                (notification, data) = cPickle.loads(chunk)
            except Exception, err:
                logging.error('Error reading message: %s', err)

            try:
                QUEUE.put_nowait((notification, data))
            except Queue.Full:
                logging.error('Queue is full, ignore message.')
            except Exception, err:
                logging.error('Error putting message: %s', err)


class MsgServer(ThreadingUnixStreamServer):

    allow_reuse_address = True

    def __init__(self, addr):
        if os.path.exists(addr):
            os.unlink(addr)
        ThreadingUnixStreamServer.__init__(self, addr, MsgHandler)

    def server_close(self):
        self.socket.shutdown(socket.SHUT_RDWR)
        self.socket.close()

    def serve_forever(self):
        """Handle one request at a time until doomsday."""
        while 1:
            if not SERVER_OWNED:
                break
            self.handle_request()


def queue_flush():
    while not QUEUE.empty():
        (notification, data) = QUEUE.get()


def worker():
    global SERVER_OWNED
    req_start_time = 0
    stats_start_time = 0
    re_connect = 0
    while SERVER_OWNED:
        (notification, data) = QUEUE.get()
        try:
            if notification.startswith('{STATS}'):
                _g_statsmgr.addAllStatisticMaps(_g_server, data)
                stats_start_time = 0
                continue
        except Exception, err:
            logging.exception('Error sending stats notification: %s', err)
            if 'Session Invalid' in str(err):
                re_connect = 1
                break
            if stats_start_time == 0:
                stats_start_time = time.time()
            if ((time.time() -  stats_start_time) > 90):
                break
            time.sleep(1)
            continue

        msg_send = False
        for i in range(5):
            try:
                _g_eventmgr.postNotificationWithData(_g_server, notification, data)
                logging.debug('Sending notification: %s', notification)
                req_start_time = 0
                msg_send = True
                break
            except Exception, err:
                logging.exception('Error sending notification: %s', err)
                if 'Session Invalid' in str(err):
                    re_connect = 1
                    break
                time.sleep(1)
        if not msg_send:
            if req_start_time == 0:
                req_start_time = time.time()
            if ((time.time() -  req_start_time) > 90):
                break
            QUEUE.put((notification, data))
            logging.debug('Queueing Failed notification: %s', notification)

        #Session Invalid error is a irrecoverable, do not wait, 
        #try to reconnect to the manager
        if re_connect == 1:
            break; 
    if QUEUE.qsize() > 1024:
        queue_flush()
        logging.debug('FLUSH QUEUE')
    SERVER_OWNED = False
    unix_server.server_close()


def delete_pidfile(pidfile):
    try:
        os.unlink(pidfile)
    except OSError, err:
        if err.errno == errno.ENOENT:
            pass
        else:
            raise


def sig_handler(signum, frame):
    global SERVER_OWNED
    SERVER_OWNED = False
    if OvmClient:
        OvmClient.OvmClient.disconnect()
    if PIDFILE != None:
        delete_pidfile(PIDFILE)
    sys.exit(0)


def receive_sigusr(signum, stack):
    if signum == signal.SIGUSR1:
        global SERVER_OWNED
        SERVER_OWNED = False
        load_core_api_bindings()


def core_api_url_components():
    """
    Get the OVM Manager Core API URL and separate out the username, password,
    and the base URL.
    """
    url = get_manager_core_api_url()
    if not url:
        raise Exception("Cannot get manager URL from agent db.")

    components = urlparse(url)
    scheme = components[0]
    netloc = components[1]
    try:
        (username_and_password, hostname_and_port) = netloc.split("@")
        (manager_un, manager_pw) = username_and_password.split(":")
    except:
        return (None, None, None)

    if not scheme or not hostname_and_port or not manager_un or not manager_pw:
        return (None, None, None)

    manager_url = scheme + "://" + hostname_and_port + "".join(components[2:])
    return (manager_url, manager_un, manager_pw)


def main():
    global PIDFILE
    PIDFILE = basename(__file__).split(".")[0]
    daemonize(PIDFILE)
    init_log()

    load_core_api_bindings()

    signal.signal(signal.SIGHUP, sig_handler)
    signal.signal(signal.SIGINT, sig_handler)
    signal.signal(signal.SIGQUIT, sig_handler)
    signal.signal(signal.SIGTERM, sig_handler)
    signal.signal(signal.SIGUSR1, receive_sigusr)

    logging.debug('NOTIFICATION SERVER STARTED')

    wrong_url = None
    while 1:
        try:
            url = get_manager_core_api_url()
            if not url or url == wrong_url:
                time.sleep(10)
                continue
            if OvmClient is None or Server is None:
                logging.error('Cannot load core api bindings.')
                time.sleep(SLEEP_ON_ERR)
                continue
            global SERVER_OWNED
            SERVER_OWNED = True
            global _g_uuid
            _g_uuid = get_server_unique_id()
            if not _g_uuid:
                raise Exception('Unable to get server UUID.')

            (host, username, password) = core_api_url_components()
            logging.debug('Trying to connect to manager.')
            cm = OvmClient.OvmClient.connect(host, username, password)
            logging.debug('Connected to manager.')
            wrong_url = None
        except Exception, err:
            logging.exception('Error initializing notification server: %s', err)
            #if 'Authentication Failed' in str(err) or 'Invalid URL Request' in str(err):
            if 'Authentication Failed' in str(err):
                wrong_url = url
                logging.debug('Wrong URL found: %s', url)
            else:
                wrong_url = None
            time.sleep(SLEEP_ON_ERR)
            continue

        try:
            foundry = cm.getFoundryContext()
            global _g_eventmgr
            global _g_server
            global _g_statsmgr
            _g_eventmgr = foundry.getEventManager()
            _g_statsmgr = foundry.getStatisticManager()
            _g_server = cm.findObject(Server.Server, _g_uuid)
            if not _g_server:
                logging.error('No manager Core API server object for %s.' % _g_uuid)
                OvmClient.OvmClient.disconnect()
                time.sleep(SLEEP_ON_ERR)
                continue

            threading.Thread(target=worker).start()
            logging.debug('Service started.')

            global unix_server
            unix_server = MsgServer(NOTIFICATION_SERVER_ADDRESS)
            unix_server.serve_forever()

            # Clean up
            unix_server.server_close()
            OvmClient.OvmClient.disconnect()
            logging.debug('Service stopped.')
        except Exception, err:
            logging.error('Error in NotificationServer process: %s', err)
            try:
                OvmClient.OvmClient.disconnect()
            except Exception, e:
                logging.error('Error disconnect: %s', e)
            time.sleep(20)
            continue


if __name__ == '__main__':
    main()
