# Copyright (C) 2006-2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

from os import listdir, makedirs, remove, rmdir, statvfs
from os.path import join, basename, exists, dirname
import string
import logging
import shutil

from OVSCommons import run_cmd, retry_cmd
from OVSConstants import *
from OVSOCFS2 import ocfs2_file_lookup, ocfs2_file_read, ocfs2_file_write, ocfs2_file_remove


def rename_file(src, dst):
    if exists(dst):
        remove(dst)
    if exists(src):
        shutil.move(src, dst)


def create_poolfs_image(poolfs_img):
    """
    Create a pool filesystem image.

    @param      poolfs_img - pool filesystem image

    @return     None. Raises an exception on failure.
    """
    if exists(poolfs_img):
        remove(poolfs_img)

    stat = statvfs(dirname(poolfs_img))
    free_space = stat.f_bsize * stat.f_bfree / 1024 / 1024
    if free_space < POOLFS_IMG_SIZE + 1:
        raise Exception("Not enough space for pool filesystem.")

    cmd = [ "dd", "if=/dev/zero", "of=%s" % poolfs_img,
            "bs=1M", "seek=%s" % POOLFS_IMG_SIZE, "count=0" ]
    run_cmd(cmd)


def make_pool_filesystem(device, poolfs_uuid, cluster_name,
        block_size="4K", cluster_size="4K", journal_size="4M"):
    """
    Make an ocfs2 filesystem for global heartbeat.

    @param      device - target device
    @param      poolfs_uuid - pool filesystem uuid
    @param      cluster_name - cluster name
    @param      block_size - block size
    @param      cluster_size - cluster size
    @param      journal size - journal size

    @return     None. Raises an exception on failure.
    """
    cmd = [ "mkfs.ocfs2",
            "-b", block_size,
            "-C", cluster_size,
            "-J", "size=%s" % journal_size,
            "-N", POOLFS_NODE_SLOTS,
            "-L", POOLFS_LABEL,
            "-U", poolfs_uuid,
            "--global-heartbeat",
            "--cluster-stack=o2cb",
            "--cluster-name=%s" % cluster_name,
            "--force",
            device ]
    run_cmd(cmd)


def create_dmnfs_device(poolfs_img, device_name):
    """
    Create a logical device for the pool filesystem image.

    @param      poolfs_img - pool filesystem image
    @device     device_name - device to create

    @return     None. Raises an exception on failure.
    """
    cmd = [ "dmsetup", "table", "--target", "nfs" ]
    output = run_cmd(cmd)
    logging.debug("dmsetup output: %s" % output)

    # dmnfs0: 0 65536 nfs /test/x.img 0
    for ln in output.splitlines():
        if len(ln.split(":")) == 2:
            device, table = ln.split(":")
            if device == device_name:
                if table and table.split()[3] == poolfs_img:
                    # already created
                    return
                else:
                    raise Exception("Device busy: %s" % ln)

    # 512 bytes sectors
    sectors = POOLFS_IMG_SIZE * 1024 * 2
    cmd = [ "dmsetup", "create", device_name,
            "--table", "0 %s nfs %s 0" % (sectors, poolfs_img) ]
    run_cmd(cmd)


def remove_dmnfs_device(device_name):
    """
    Remove a dmnfs device.

    @param      device_name - device to remove

    @return     None. Raise an exception on failure.
    """
    cmd = [ "dmsetup", "table", "--target", "nfs" ]
    output = run_cmd(cmd)
    logging.debug("dmsetup output: %s" % output)

    # dmnfs0: 0 65536 nfs /test/x.img 0
    for ln in output.splitlines():
        device, table = ln.split(":")
        if device == device_name:
            cmd = [ "dmsetup", "remove", device_name ]
            run_cmd(cmd)


def mount_device(device, mount_point):
    """
    Mount a device.

    @param      device - device to mount
    @param      mount_point - mount point

    @return     None. Raise an exception on failure.
    """
    logging.debug("Trying to mount %s to %s", device, mount_point)

    if not exists(mount_point):
        makedirs(mount_point)

    cmd = [ "cat", "/proc/mounts" ]
    output = run_cmd(cmd)

    mounted = False
    for ln in output.splitlines():
        if device == ln.split()[0]:
            mounted = True
            break

    if mounted:
        mnt = ln.split()[1]
        if mnt != mount_point:
            raise Exception("%s is already mounted on %s" % (device, mnt))
        cmd = [ "mount", "-o", "remount", device ]
        retry_cmd(cmd)
        logging.debug("%s re-mounted to %s", device, mount_point)
    else:
        cmd = [ "mount", device, mount_point ]
        retry_cmd(cmd)
        logging.debug("%s mounted to %s", device, mount_point)


def umount_device(device):
    """
    Umount a device.

    @param      device - device to umount

    @return     None. Raise an exception on failure.
    """
    logging.debug("Trying to umount %s", device)

    mount_point = None
    for ln in open("/proc/mounts", "r"):
        parts = ln.split()
        if device == parts[0] or device == parts[1]:
            mount_point = parts[1]
            break

    if mount_point:
        # force unmount, in case of an unreachable NFS system
        cmd = [ "umount", "-f", device ]
        try:
            retry_cmd(cmd)
            try:
                # remove mount point if empty
                rmdir(mount_point)
            except Exception, e:
                logging.warn("Error removing mount point: %s")
        except Exception, e:
            # list all processes accessing files on the filesystem
            cmd = [ "lsof", mount_point ]
            try:
                output = run_cmd(cmd).strip()
                logging.error("Device or resource busy: \n%s", output)
            except:
                # unexpected dmnfs device?
                cmd = [ "dmsetup", "table", "--target", "nfs" ]
                try:
                    output = run_cmd(cmd)
                    for line in output.splitlines():
                        if mount_point in line:
                            logging.error("Unexpected dmnfs device: %s", line)
                except:
                    logging.exception("Error getting dmnfs devices.")
            # raise umount error
            raise e

    logging.debug("%s umounted", device)


def parse_metafile(contents):
    """
    Parse metafile contents to a dict.

    @param      contents - metafile contents as a string

    @return     Metafile contents as a dict.
    """
    d = {}

    for line in contents.splitlines():
        # deal with comments
        ln = line.split("#")[0].strip()
        # parse key/value pairs
        if ln == "":
            continue
        elif "=" in ln and len(ln.split("=")) == 2:
            key, value = map(string.strip, ln.split("="))
            d[key] = value
        else:
            raise Exception("Invalid line in metafile: %s" % line)

    return d


def ocfs2_get_meta_data(device, filename):
    """
    @return     Meta data as a dict.
                Raises an exception on failure.
    """
    if ocfs2_file_lookup(device, filename) == 0:
        contents = ocfs2_file_read(device, filename)
        return parse_metafile(contents)
    else:
        return {}


def nfs_create_pool_filesystem(poolfs_target, cluster_name, poolfs_uuid,
        poolfs_nfsbase_uuid, manager_uuid, pool_uuid):
    # mount NFS base
    mount_point = join(NFS_MOUNT, poolfs_nfsbase_uuid)
    mount_device(poolfs_target, mount_point)
    # make pool filesystem
    poolfs_img = join(mount_point, POOLFS_IMG)
    create_poolfs_image(poolfs_img)
    make_pool_filesystem(poolfs_img, poolfs_uuid, cluster_name)
    # generate poolfs metafile contents
    poolfs_meta = join(mount_point, POOLFS_META)
    contents = """OVS_POOLFS_UUID=%s
OVS_POOLFS_NFSBASE_UUID=%s
OVS_POOLFS_MGR_UUID=%s
OVS_POOLFS_VERSION=%s
OVS_POOLFS_POOL_UUID=%s
""" % (poolfs_uuid,
poolfs_nfsbase_uuid,
manager_uuid,
POOLFS_VERSION,
pool_uuid,
)
    try:
        # write poolfs metafile
        f = file(poolfs_meta, "w")
        try:
            f.write(contents)
        finally:
            f.close()
        # umount poolfs nfsbase
        umount_device(poolfs_target)
    except:
        if exists(poolfs_meta):
            remove(poolfs_meta)


def lun_create_pool_filesystem(poolfs_target, cluster_name, poolfs_uuid,
        manager_uuid, pool_uuid):
    # destroy pool filesystem
    lun_destroy_pool_filesystem(poolfs_target)
    # make pool filesystem
    make_pool_filesystem(poolfs_target, poolfs_uuid, cluster_name)
    # generate poolfs metafile contents
    contents = """OVS_POOLFS_UUID=%s
OVS_POOLFS_MGR_UUID=%s
OVS_POOLFS_VERSION=%s
OVS_POOLFS_POOL_UUID=%s
OVS_POOLFS_LUN_UUID=%s
""" % (poolfs_uuid,
manager_uuid,
POOLFS_VERSION,
pool_uuid,
basename(poolfs_target),
)
    try:
        # write poolfs metafile
        ocfs2_file_write(poolfs_target, POOLFS_META, contents)
    except:
        lun_destroy_pool_filesystem(poolfs_target)


def create_pool_filesystem(poolfs_type, poolfs_target, cluster_name,
        poolfs_uuid, poolfs_nfsbase_uuid, manager_uuid, pool_uuid,
        block_size="4K", cluster_size="4K", journal_size="4M"):
    """
    Create a pool filesystem.

    @param      poolfs_type - "nfs" or "lun"
    @param      poolfs_target -
                physical location of the pool filesystem
    @param      cluster_name - cluster to use the pool filesystem
    @param      poolfs_uuid -
                a unique identifier for the pool filesystem
    @param      poolfs_nfsbase_uuid -
                a unique identifier for the base NFS filesystem
    @param      manager_uuid - a unique identifier for the manager
    @param      pool_uuid - a unique identifier for the server pool
    @param      block_size - mkfs.ocfs2 option
    @param      cluster_size - mkfs.ocfs2 option
    @param      journal_size - mkfs.ocfs2 option

    @return     None. Raises an exception on failure.
    """
    if poolfs_type == "nfs":
        nfs_create_pool_filesystem(poolfs_target, cluster_name, poolfs_uuid,
                poolfs_nfsbase_uuid, manager_uuid, pool_uuid)
    elif poolfs_type == "lun":
        lun_create_pool_filesystem(poolfs_target, cluster_name, poolfs_uuid,
                manager_uuid, pool_uuid)
    else:
        raise ValueError("Invalid pool filesystem type: %s" % poolfs_type)


def nfs_destroy_pool_filesystem(poolfs_target, poolfs_nfsbase_uuid):
    mount_point = join(NFS_MOUNT, poolfs_nfsbase_uuid)
    mount_device(poolfs_target, mount_point)
    poolfs_meta = join(mount_point, POOLFS_META)
    if exists(poolfs_meta):
        remove(poolfs_meta)
    umount_device(poolfs_target)


def lun_destroy_pool_filesystem(poolfs_target):
    cmd = [ "dd", "if=/dev/zero", "of=%s" % poolfs_target,
            "bs=1M", "count=4" ]
    run_cmd(cmd)


def destroy_pool_filesystem(poolfs_type, poolfs_target, poolfs_uuid,
                                                poolfs_nfsbase_uuid):
    """
    Destroy a pool filesystem.

    @param      poolfs_type - "nfs" or "lun"
    @param      poolfs_target -
                physical location of the pool filesystem
    @param      poolfs_uuid -
                a unique identifier for the pool filesystem
    @param      poolfs_nfsbase_uuid -
                a unique identifier for the base NFS filesystem

    @return     None. Raises an exception on failure.
    """
    if poolfs_type == "nfs":
        nfs_destroy_pool_filesystem(poolfs_target, poolfs_nfsbase_uuid)
    elif poolfs_type == "lun":
        lun_destroy_pool_filesystem(poolfs_target)
    else:
        raise ValueError("Invalid pool filesystem type: %s" % poolfs_type)
