#!/usr/bin/env python

# Copyright (C) 2006-2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import logging
from time import sleep
from os.path import join, basename

from OVSLog import init_log
from OVSDaemon import daemonize
from OVSConfig import get_config
from OVSAgentDB import read_item, write_item, dump_db, get_cluster_db_home
from OVSServerPool import is_clustered
from OVSConstants import *
from OVSCommons import host_alive
from OVSMasterFailover import (acquire_master_dlm_lock,
        release_master_dlm_lock, is_holding_vip, is_master,
        ifconfig_vip_up, ifconfig_vip_down, setup_master_env)
from OVSNotificationProxy import notify_manager
from utils import filelock


def send_remaster_event():
    hostname = read_item(LOCAL_SERVER_DB, "registered_hostname")
    msg = "{CLUSTER} {REMASTER} master is [%s]" % hostname
    try:
        notify_manager(msg)
    except Exception, e:
        logging.debug("Error sending notification: %s", e)


take_crown = True
def cluster_remaster():
    global take_crown
    db_home = get_cluster_db_home()
    pool_vip = read_item("server_pool", "pool_virtual_ip", db_home)
    if acquire_master_dlm_lock():
        if pool_vip:
            if host_alive(pool_vip):
                if is_holding_vip(pool_vip):
                    if is_master():
                        write_item(LOCAL_SERVER_DB, "is_master", True)
                        if take_crown:
                            take_crown = False
                            logging.info("It is master, VIP=%s" % pool_vip)
                            send_remaster_event()
                    else:
                        take_crown = True
                        ifconfig_vip_down(pool_vip)
                        write_item(LOCAL_SERVER_DB, "is_master", False)
                        release_master_dlm_lock()
                        logging.debug("It is not master: VIP released")
                else:
                    take_crown = True
                    write_item(LOCAL_SERVER_DB, "is_master", False)
                    release_master_dlm_lock()
                    logging.debug("VIP in use: master lock released")
            else:
                take_crown = True
                try:
                    setup_master_env()
                    logging.debug("Master server elected: setup done")
                    write_item(LOCAL_SERVER_DB, "is_master", True)
                    send_remaster_event()
                finally:
                    if not is_master():
                        release_master_dlm_lock()
    else:
        take_crown = True
        write_item(LOCAL_SERVER_DB, "is_master", False)


def remaster():
    d = dump_db(LOCAL_SERVER_DB)
    roles = d.get("roles")
    if roles and "master" in roles:
        lock = filelock.FileLock(CLUSTER_LOCK)
        lock.acquire(wait=60)
        try:
            if is_clustered():
                cluster_remaster()
            else:
                vip = d.get("pool_virtual_ip")
                ip = d.get("registered_ip")
                master_flag = d.get("is_master")
                if master_flag and vip and not is_holding_vip(vip):
                    if host_alive(vip):
                            logging.debug("VIP in use: conflict detected")
                    else:
                        ifconfig_vip_up(vip, ip)
                        logging.info("This node is the master, VIP=%s" % vip)
        finally:
            lock.release()


def run():
    logging.info("REMASTER SERVER STARTED")

    config = get_config()
    remaster_interval = int(config.get("remaster", "interval"))

    while True:
        try:
            sleep(remaster_interval)
            remaster()
        except Exception, e:
            logging.exception("Error in remaster process: %s", e)
            sleep(SLEEP_ON_ERR)


def main():
    pid_filename = basename(__file__).split(".")[0]
    daemonize(pid_filename)
    init_log()
    try:
        run()
    except Exception, e:
        logging.exception("Error in remaster process: %s", e)


if __name__ == "__main__":
    main()
