# Copyright (C) 2006-2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import os
import codecs
import socket
import logging
from OVSCommons import run_cmd


INTERFACE_CONFIG_PREFIX = "/etc/sysconfig/network-scripts/"


def ovs_br_config(action, br_name, net_dev):
    """
    Configure a Standard Bridge.

    @param action (start | stop)
          start: Creates the bridge, Copies the IP and MAC addresses from netdev to bridge,
                 enslaves net_dev to bridge, and saves bridge config files in
                 /etc/sysconfig/network-scripts
          stop:  Removes net_dev from the bridge,transfers addresses, routes from bridge to
                 net_dev, deletes the bridge, and revomes bridge config files from
                 /etc/sysconfig/network-scripts

    @param br_name       The bridge name to add
    @param net_dev       The physical interface to add to the bridge

    @return    If successful, returns the names of the bridge and it's physical interface.
               Raises an exception on failure
    Restriction:
    -net_dev  must be physical, or bond
    -net_dev  must not be member of a bridge, or slave to a bond
    """
    if action != "start" and action != "stop":
        raise Exception ("action %s is not supported" % action)
    if br_name.startswith("bond") or br_name.startswith("eth"):
        raise Exception ("bridge name can not be prefixed with (bond or eth) %s" % br_name)
    cmd = ["/etc/xen/scripts/ovs-network-bridge", "%s" % action, "bridge=%s" % br_name, "netdev=%s" % net_dev]
    run_cmd(cmd)
    logging.info("ovs-network-bridge: success. %s bridge %s on interface %s" % (action, br_name, net_dev))
    return "%s bridge=%s netdev=%s" % (action, br_name, net_dev)


def ovs_async_bridge(action, br_name, net_dev):
    """
    Configure a Standard Bridge.This command to be used in async mode.

    @param action (start | stop)
          start: Creates the bridge, Copies the IP and MAC addresses from netdev to bridge,
                 enslaves net_dev to bridge, and saves bridge config files in
                 /etc/sysconfig/network-scripts
          stop:  Removes net_dev from the bridge,transfers addresses, routes from bridge to
                 net_dev, deletes the bridge, and revomes bridge config files from
                 /etc/sysconfig/network-scripts

    @param br_name       The bridge name to add
    @param net_dev       The physical interface to add to the bridge

    @return    If successful, returns the names of the bridge and it's physical interface.
               Raises an exception on failure
    Restriction:
    -net_dev  must be physical, or bond
    -net_dev  must not be member of a bridge, or slave to a bond
    """
    if action != "start" and action != "stop":
        raise Exception ("action %s is not supported" % action)
    if br_name.startswith("bond") or br_name.startswith("eth"):
        raise Exception ("bridge name can not be prefixed with (bond or eth) %s" % br_name)
    cmd = ["/etc/xen/scripts/ovs-network-bridge", "%s" % action, "bridge=%s" % br_name, "netdev=%s" % net_dev]
    run_cmd(cmd)


def ovs_vlan_bridge(action, br_name, net_dev, vlan_id):
    """
    Configure a VLAN Bridge.

    @param action (start | stop)
          start: Creates the bridge, creats VLAN on net_dev,enslaves
                 the VLAN to the bridge, and saves VLAN bridge config
                 files in /etc/sysconfig/network-scripts
          stop:  Removes the VLAN from the bridge, removes the VLAN,
                 deletes the bridge, and removes VLAN bridge config files
                 from /etc/sysconfig/network-scripts

    @param br_name       The bridge name to add
    @param net_dev       The physical interface on which to create a VLAN.
    @param vlan_id       VLAN ID (1-4095). VLAN ID of 1 is the untagged VLAN.

    @return       If successful, returns the names of the bridge and it's VLAN interface
                  Raises an exception on failure
    Restriction:
    -net_dev  must be physical, or bond
    -net_dev  must not be member of a bridge, or slave to a bond
    -VLAN ID must not exist on the same interface
    """
    if action != "start" and action != "stop":
        raise Exception ("action %s is not supported" % action)
    if br_name.startswith("bond") or br_name.startswith("eth"):
        raise Exception ("bridge name can not be prefixed with (bond or eth) %s" % br_name)
    cmd = ["/etc/xen/scripts/ovs-vlan-bridge", "%s" % action, "bridge=%s" % br_name, "vlanid=%d" % int(vlan_id), "netdev=%s" % net_dev]
    run_cmd(cmd)
    logging.info("ovs-vlan-bridge: success. %s bridge %s on interface %s vlan %d" % (action, br_name, net_dev, int(vlan_id)))
    return "%s bridge=%s netdev=%s vlan %d" % (action, br_name, net_dev, int(vlan_id))


def ovs_bond_config(action, bond, *args):
    """
    Configure Bonding.

    @param action (start | stop)
          start: Bond two physical interfaces to one logical interface.
                 (eth0,eth1 --> bond0), and saves bond config files in
                 /etc/sysconfig/network-scripts
          stop:  Removes channel bond relation, removes bond config files from
                 /etc/sysconfig/network-scripts

    @param bond   One of the logical channel bonds (bond0, bond1 ...etc)
    @param args   Is a list of the following arguments: (Not required when removing a bond)

          bondmode   Current supported bonding modes (1 = active-backup, 4 = 802.3ad
                     6 = balance-alb).
          primary    Primary physical interface (eth0,eth1 ..etc).
          backup     Backup physical interface  (eth0,eth1 ..etc)
                     Primary/backup are only valid for active-backup mode,
                     they are used as a place holder when configuring the other two modes.

    @return       If successful, returns the names of the physical interfaces and the bond
                  Raises an exception on failure
    Restriction:
    -bond     must be one of the logical channel bonds (bond0, bond1 ...etc)
    -primary  must be physical, and not be member of a bridge, or slave to a bond
    -backup   must be physical, and not be member of a bridge, or slave to a bond

    """
    if action != "start" and action != "stop":
        raise Exception ("action %s is not supported" % action)
    if action == "start":
        if len(args) < 3:
            raise Exception ("ovs-bond-if failed: bond mode required")
        mode = args[0]
        primary = args[1]
        backup = args[2]
        cmd = ["/etc/xen/scripts/ovs-bond-if", "%s" % action, "bondm=%s" % bond, "bondmode=%d" % int(mode), "primary=%s" % primary, "backup=%s" % backup]
    else:
        cmd = ["/etc/xen/scripts/ovs-bond-if", "%s" % action, "bondm=%s" % bond]
    run_cmd(cmd)
    if action == "start":
        logging.info("ovs-bond-if: success. %s bonding interfaces %s %s" % (action, primary, backup))
        return "%s bonding inetrfaces %s %s" % (action, primary, backup)
    else:
        logging.info("ovs-bond-if: success. removed %s" % (bond))
        return "removed bond %s" % (bond)

def ovs_async_bond(action, bond, *args):
    """
    Configure Bonding. This command to be used in async mode.

    @param action (start | stop)
          start: Bond two physical interfaces to one logical interface.
                 (eth0,eth1 --> bond0), and saves bond config files in
                 /etc/sysconfig/network-scripts
          stop:  Removes channel bond relation, removes bond config files from
                 /etc/sysconfig/network-scripts

    @param bond   One of the logical channel bonds (bond0, bond1 ...etc)
    @param args   Is a list of the following arguments: (Not required when removing a bond)

          bondmode   Current supported bonding modes (1 = active-backup, 4 = 802.3ad
                     6 = balance-alb).
          primary    Primary physical interface (eth0,eth1 ..etc).
          backup     Backup physical interface  (eth0,eth1 ..etc)
                     Primary/backup are only valid for active-backup mode,
                     they are used as a place holder when configuring the other two modes.

    @return       If successful, returns the names of the physical interfaces and the bond
                  Raises an exception on failure
    Restriction:
    -bond     must be one of the logical channel bonds (bond0, bond1 ...etc)
    -primary  must be physical, and not be member of a bridge, or slave to a bond
    -backup   must be physical, and not be member of a bridge, or slave to a bond

    """
    if action != "start" and action != "stop":
        raise Exception ("action %s is not supported" % action)
    if action == "start":
        if len(args) < 3:
            raise Exception ("ovs-bond-if failed: bond mode required")
        mode = args[0]
        primary = args[1]
        backup = args[2]
        cmd = ["/etc/xen/scripts/ovs-bond-if", "%s" % action, "bondm=%s" % bond, "bondmode=%d" % int(mode), "primary=%s" % primary, "backup=%s" % backup]
    else:
        cmd = ["/etc/xen/scripts/ovs-bond-if", "%s" % action, "bondm=%s" % bond]
    run_cmd(cmd)


def ovs_bond_op(action, bond, backup):
    """
    Configure Bonding.

    @param action (add | remove)
          add: Adds a physical interfaces to an existing bond.
                 (eth0,eth1 --> bond0), and saves bond config files in
                 /etc/sysconfig/network-scripts
          remove:  Removes an interfacse from an existing bond and restore the
                   interface config file to /etc/sysconfig/network-scripts

    @param bond   One of the logical channel bonds (bond0, bond1 ...etc)
    @param backup  Physical interface to add/remove (eth0, eth1 ...etc)

    @return       If successful, returns the names of the physical interfaces and the bond
                  Raises an exception on failure
    Restriction:
    -bond     must be one of the logical channel bonds (bond0, bond1 ...etc)
    -backup   must be physical, and not be member of a bridge, or slave to a bond

    """
    if action != "add" and action != "remove":
        raise Exception ("action %s is not supported" % action)
    cmd = ["/etc/xen/scripts/ovs-bond-op", "%s" % action, "bondm=%s" % bond, "backup=%s" % backup]
    run_cmd(cmd)
    if action == "add":
        logging.info("ovs-bond-op: success. added interface %s to %s" % (backup,bond))
        return "added inetrface %s to %s" % (backup,bond)
    else:
        logging.info("ovs-bond-op: success. removed interface %s from %s" % (backup,bond))
        return "removed interface %s form %s" % (backup,bond)

def ovs_bond_mode(bond, mode):
    """
    Change Bond mode.

    @param bond   One of the logical channel bonds (bond0, bond1 ...etc)
    @mode         Current supported bonding modes (1 = active-backup, 4 = 802.3ad
                  6 = balance-alb).
    @return       If successful, returns bond's names and its new mode
                  Raises an exception on failure
    Restriction:
    -bond     must be one of the logical channel bonds (bond0, bond1 ...etc)

    """
    if not bond:
        raise Exception ("bond name is required")
    if not mode:
        raise Exception ("bond mode (1,4,6) is required")
    cmd = ["/etc/xen/scripts/ovs-bond-mode", "bondm=%s" % bond, "bondmode=%d" % int(mode)]
    run_cmd(cmd)
    logging.info("ovs-bond-mode: success. %s mode changed to %s" % (bond,mode))
    return "%s mode changed to %s" % (bond,mode)

def ovs_local_config(action, br_name ):
    """
    Configure a Local Bridge ((NIC-less bridge)

    @param action (start | stop)
         start: Creates local bridge without a physical interface, and saves bridge config
                file in /etc/sysconfig/network-scripts
          stop: Deletes local bridge, removes bridge config file from
                /etc/sysconfig/network-scripts

    @param br_name       The bridge name to add

    @return       If successful, returns the name of the bridge
                  Raises an exception on failure
    """
    if action != "start" and action != "stop":
        raise Exception ("action %s is not supported" % action)
    cmd = ["/etc/xen/scripts/ovs-local-bridge", "%s" % action, "bridge=%s" % br_name]
    run_cmd(cmd)
    logging.info("ovs-local-bridge: success. %s local bridge %s" % (action, br_name))
    return "%s local bridge %s" % (action, br_name)


def ovs_vlan_config(action, interface, vlanid ):
    """
    Creates a VLAN interface on a physical, or a bond interface.

    @param  action (add|remove)
          add:    Creates a VLAN on an interface,saves the VLAN config file in
                  /etc/sysconfig/network-scripts
          remove: Removes a VLAN from an interfacei,removes its config file from
                  /etc/sysconfig/network-scripts

    @param   interface    The interface on which to create a VLAN
    @param   vlanid       VLAN ID (2-4095)

    @return  If successful, returns the interface, and VLAN created
             Raises an exception on failure

    Restriction:
    -Interface  must be physical, or bond
    -Interface must not be member of a bridge, or slave to a bond
    -VLAN ID must not exist on the same interface
    """

    if action != "add" and action != "remove":
        raise Exception ("action %s is not supported" % action)
    cmd = ["/etc/xen/scripts/ovs-vlan-config", "%s" % action, "%s" % interface, "%d" % int(vlanid)]
    run_cmd(cmd)
    logging.info("ovs-vlan-config: success. %s VLAN %d to %s" % (action, int(vlanid), interface))
    return "%s VLAN %d interface %s" % (action, int(vlanid), interface)


def ovs_ip_config(interface, optype, *args):
    """
    Assigns/flushes IP, netmask address to a physical,VLAN, and bond interfaces.

    @param  interface    The interface on which to assign
    @param  optype (static|dynamic|flush)
        static:  Assigns the given IP, and netmask to the interface, and
                     saves the config file to /etc/sysconfig/network-scripts.
        dynamic: Flushes current address, and creats and save the config
                     file to /etc/sysconfig/network-scripts, (BOOTPROTO=dhcp)
        flush:   flushes the interface address,routes, removes the current
                     config file from /etc/sysconfig/network-scripts.
                     Creats a new one with BOOTPROTO=static

    @args   Required for the static option, otherwise it is ignored
        IP address:  IPv4 address in decimal notation (101.230.112)
        netmask:     Standard netmask in a decimal notation,NOT CIDR.
                         example(255.255.255.0)

    @return  If successful, returns the interface, and addresses added/flushed
             Raises an exception on failure

    Restriction:
    -Interface must be physical, VLAN, or a Bond
    -Interface must not be a bridge port, or slave to a bond
    -Addresses must be valid in a decimal notation
    """
    if optype != "static" and optype != "dynamic" and optype != "flush":
        raise Exception ("operation type %s is not supported" % optype )
    if optype == "static":
        if len(args) < 2:
            raise Exception ("ovs-ip-config faild: IP Adress and netmask are required")
        ipaddr = args[0]
        netmask = args[1]

        #Validate IP
        try:
            socket.inet_aton(ipaddr)
            seprator = ipaddr.split(".")
        except:
            raise Exception ("IP %s is not valid." % ipaddr)
        if len(seprator) != 4:
            raise Exception ("ovs-ip-config faild: IP %s is not valid." % ipaddr)

        #Validate netmask
        seprator = netmask.split(".")
        if len(seprator) != 4:
            raise Exception ("ovs-ip-config faild: netmask %s is not legal valid." % netmask )

        cmd = ["/etc/xen/scripts/ovs-ip-config", "%s" % interface, "%s" % optype, "%s" % ipaddr, "%s" % netmask]
    else:
        cmd = ["/etc/xen/scripts/ovs-ip-config", "%s" % interface, "%s" % optype]

    run_cmd(cmd)
    if optype == "static":
        logging.info("ovs-ip-config: success: interface %s  IP %s mask %s" % (interface, ipaddr, netmask))
        return "interface %s  IP %s mask %s" % (interface, ipaddr, netmask)
    else:
        logging.info("ovs-ip-config: success: interface %s %s" % (interface, optype))
        return "interface %s %s" % (interface, optype)


def ovs_if_meta(interface, data):
    """
    This function creates meta data file meta-interface, and write the string
    (METADATA=data) to it.  This string is used by the manager to identify
    networks that interfaces belong to. Dom0 does not make used of this string,
    it just saves it and returns it during running, saved network discovery.

     - If an interface already has a meta data string, then it gets
       replace by the new one
     - An empty meta data string, indicates to remove the existing string
       (remove the meta-interface) file

    @param  interface    physical,VLAN, bond ...etc interfaces
    @param  data         meta data to save

    @return  If successful, returns the interface, and meta data
             Raises an exception on failure

    Restriction:
    - data string that starts with leading spaces will be rejected
    """
    if not os.path.exists('/sys/class/net/%s' % (interface)):
        raise Exception("ovs_set_metadata: interface /sys/class/net/%s does not exists" % (interface))

    if data.startswith(" "):
        raise Exception("ovs_set_metadata: meta data must not start with leading white space")

    cfg_file = "%smeta-%s" % (INTERFACE_CONFIG_PREFIX, interface)
    if data == '': #remove meta data file
        if not os.path.exists(cfg_file):
            raise Exception ("file does not exist: %s" % (cfg_file))
        os.remove(cfg_file)
        logging.info("ovs_set_metadata: success: metat file %s is removed" % (cfg_file))
        return "metat file %s is removed" % (cfg_file)

    # always over write exisiting data
    f = codecs.open(cfg_file, "w", encoding="utf-8")
    f.write("#This file was dynamically created by OVM manager. Please Do not edit\n")
    f.write("METADATA=%s\n" % (data))
    f.close()
    logging.info("ovs_set_metadata: success: set metatdata %s for interface %s" % (data, interface))
    return "set meta data %s for interface %s" % (data, interface)


def ovs_change_mtu(interface, mtu):
    """
    Change MTU on a physical, or a bond interface, including bridges,
    and VLANs


    @param   interface    Physical, bond interface, local bridge
    @param   MTU          Values are 1500 or 9000

    @return  If successful, returns the interface, and the MTU Value set to
             Raises an exception on failure

    Restriction:
    -Interface  must be physical, or bond
    -MTU Value must be 1500 or 9000
    """

    if mtu != "1500" and mtu != "9000":
        raise Exception ("MTU %s is out of range. MTU values must be (1500 | 9000)" % mtu)
    cmd = ["/etc/xen/scripts/ovs-change-mtu", "netdev=%s" % interface, "mtu=%d" % int(mtu)]
    run_cmd(cmd)
    logging.info("ovs-change-mtu: success. interface %s set to %s" % (interface, mtu))
    return "interface %s set to %s" % (interface, mtu)
