# Copyright (C) 2006-2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import socket
import time

from OVSCommons import old_run_cmd, host_alive
from OVSConstants import *
from OVSAgentDB import write_item, read_item, update_db, get_cluster_db_home
from OVSServerPool import is_clustered
from OVSMasterFailover import (ifconfig_vip_up, ifconfig_vip_down,
        is_holding_vip, acquire_master_dlm_lock, release_master_dlm_lock)
from utils import filelock


def validate_cluster_vip(vip, base_ip):
    """
    Validate virtual IP for the cluster:
    1. vip and base_ip must be valid
    2. vip should not be in use
    3. vip and base_ip are in the same network
    """
    def _legal_ipaddr(ip):
        try:
            socket.inet_aton(ip)
            return True
        except:
            return False

    def _get_network(ip, mask):
        cmd = "ipcalc -n %s %s" % (ip, mask)
        output = old_run_cmd(cmd).strip()
        return output[8:]

    if not _legal_ipaddr(vip):
        raise Exception("Invalid VIP address: %s" % vip)
    if not _legal_ipaddr(base_ip):
        raise Exception("Invalid base IP address: %s" % base_ip)

    if host_alive(vip):
        if is_holding_vip(vip):
            return
        else:
            raise Exception("VIP in use: %s" % vip)

    cmd = "ifconfig | grep '%s ' | awk -F: '{print $4}'" % base_ip
    netmask = old_run_cmd(cmd).strip()
    if netmask == "":
        raise Exception("Failed to get netmask for base IP: %s" % base_ip)
    else:
        netmask = netmask.splitlines()[-1]

    network1 = _get_network(base_ip, netmask)
    network2 = _get_network(vip, netmask)
    if network1 != network2:
        raise Exception("VIP %s and base IP %s are not in the same network."
                                                        % (vip, base_ip))


def configure_virtual_ip(virtual_ip, base_ip):
    """
    Configure pool virtual IP.

    @param      virutal ip - pool virtual ip
    @param      base_ip - ip that manager uses to talk to this server

    @return     None. Raises an exception on failure.
    """
    lock = filelock.FileLock(CLUSTER_LOCK)
    lock.acquire(wait=60)
    try:
        validate_cluster_vip(virtual_ip, base_ip)
        ifconfig_vip_up(virtual_ip, base_ip)
        write_item(LOCAL_SERVER_DB, "is_master", True)
        if is_clustered():
            my_hostname = read_item(LOCAL_SERVER_DB, "registered_hostname")
            master_info = {
                "pool_master_hostname": my_hostname,
                "pool_virtual_ip": virtual_ip}
            update_db("server_pool", master_info, get_cluster_db_home())
    finally:
        lock.release()


def deconfigure_virtual_ip(virtual_ip):
    """
    Deconfigure pool virtual IP.

    @param      virutal ip - pool virtual ip

    @return     None. Raises an exception on failure.
    """
    lock = filelock.FileLock(CLUSTER_LOCK)
    lock.acquire(wait=60)
    try:
        ifconfig_vip_down(virtual_ip)
        write_item(LOCAL_SERVER_DB, "is_master", False)
        if is_clustered():
            master_info = {
                "pool_master_hostname": None,
                "pool_virtual_ip": None}
            update_db("server_pool", master_info, get_cluster_db_home())
    finally:
        lock.release()
