# Copyright (C) 2006-2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import sys
import logging

from OVSDiscoverCommon import SimpleDiscoveryXML, DiscoveryLock

def discover_storage_plugins():
    """
    Discover the installed Storage Connect plugins on this OVM Server.

    @return: None. Raises an exception on failure.
    """
    lock = DiscoveryLock("discover_storage_plugins")
    if not lock.acquire():
        raise Exception("discover_storage_plugins already in progress")

    class DiscoverPlugins(object):
        """
        Discover the installed plugins.
        """
        def __init__(self):
            """
            Constructor for the Storage Connect Plugin Discovery.
            """
            self.__plugin_mgr__ = PluginManager(logging.getLogger())
            self.__plugin_names__ = None

        def getInfoOnPlugins(self):
            """
            Obtain information on all the plugins installed in the system.

            @return: A dict with the information on all the discovered plugins.
            """
            self.__plugin_names__ = self.__plugin_mgr__.discoverPlugins(ignore_failed = True)
            info = {}
            for plugin_name in self.__plugin_names__:
                p_info = self.__plugin_mgr__.getPluginInfo(plugin_name)
                abilities = p_info['plugin_ability']
                for ability in abilities.keys():
                    abilities[ability] = OSCPlugin.ABILITY_TYPES.getID(abilities[ability])

                info[plugin_name] = p_info

            return info

        def genDiscoverResult(self, doc):
            """
            Generate the discover request result XML for the plugins.

            @param:  XML Document to insert information about the plugins into.

            @return: None. Raises an exception on error.
            """
            if not self.__plugin_names__ or len(self.__plugin_names__) <= 0:
                self.__plugin_names__ = self.__plugin_mgr__.discoverPlugins(ignore_failed = True)
                if not self.__plugin_names__ or len(self.__plugin_names__) <= 0:
                    # We expect at least one plugin (the generic ones) to be discovered
                    raise Exception("Unable to obtain Storage Connect plugin list")

            doc.push("storage_plugin_info_list")
            for impl_name in self.__plugin_names__:
                try:
                    plugin_info = self.__plugin_mgr__.getPluginInfo(impl_name)

                except:
                    # We just ignore any plugin that gives any kind of error
                    continue

                doc.push("storage_plugin_info", "plugin_impl_name", impl_name)
                for info_key in plugin_info.keys():
                    if info_key != "plugin_ability":
                        if type(plugin_info[info_key]) == str:
                            doc.add(info_key, plugin_info[info_key])

                        elif type(plugin_info[info_key]) == list:
                            doc.add(info_key, ",".join(plugin_info[info_key]))

                        else:
                            doc.add(info_key, str(plugin_info[info_key]))

                abilities = plugin_info.get("plugin_ability", None)
                if abilities:
                    doc.push("abilities")
                    for ability in abilities.keys():
                        if ability == "max_access_entries":
                            doc.add(ability, str(abilities[ability]))

                        elif ability == "backing_device_type":
                            abilities[ability] = OSCPlugin.BACKING_DEVICE_TYPES.getID(abilities[ability])
                            doc.add(ability, abilities[ability])

                        else:
                            abilities[ability] = OSCPlugin.ABILITY_TYPES.getID(abilities[ability])
                            doc.add(ability, abilities[ability])

                    doc.pop() # abilities

                doc.pop() # storage_plugin_info

            doc.pop() # storage_plugin_info_list

    try:
        if "/opt/storage-connect" not in sys.path: sys.path.append("/opt/storage-connect")

        from PluginManager import PluginManager
        import OSCPlugin

        doc = SimpleDiscoveryXML("Discover_Storage_Plugins_Result")
        plugin_discovery = DiscoverPlugins()
        plugin_discovery.genDiscoverResult(doc)

    except Exception, e:
        lock.release()
        raise Exception("Unable to discover Storage Connect plugins: %s" % e)

    try:
        result = doc.finish()

    except Exception, e:
        lock.release()
        raise Exception("Unable to format discover_plugin_info XML: %s" % e)

    lock.release()
    return result
