# Copyright (C) 2006-2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import os
import re
import sys
import copy
import glob
import random
import string
import tarfile
import logging
import tempfile
import rpm
import yum
import OVSConstants
from distutils.sysconfig import get_python_lib
from iniparse import INIConfig
from iniparse.compat import ConfigParser
from xen.xend import sxp
from OVSCommons import run_cmd
from password import update_password
from OVSDownload import ovs_download_file
from OVSConfig import reload_agent_server, reload_notification_server
from utils import filelock, xmlrpclib
from utils.dateBackend import dateBackend
from utils.timezoneBackend import timezoneBackend


xend_conf = '/etc/xen/xend-config.sxp'
yum_conf = '/etc/yum.conf'
yum_repo_conf = '/etc/yum.repos.d/ovm.repo'
yum_lock = '/var/run/ovs-agent/yum.lock'


def read_xend_cfg():
    f = open(xend_conf)
    s = sxp.parse(f)
    f.close()
    if s:
        return dict(s)
    return {}


def write_xend_cfg(cfg):
    f = open(xend_conf, 'w')
    for k, v in cfg.items():
        sxp.show([k, v], f)
        f.write('\n')
    f.close()


def restart_xend():
    run_cmd(['/sbin/service', 'xend', 'restart'])


def get_xend_config():
    lock = filelock.LockFile(xend_conf)
    lock.acquire(wait=5)
    try:
        return read_xend_cfg()
    finally:
        lock.release()


def configure_xend(params):
    lock = filelock.LockFile(xend_conf)
    lock.acquire(wait=5)
    try:
        cfg = read_xend_cfg()
        new_cfg = copy.deepcopy(cfg)
        new_cfg.update(params)
        if new_cfg != cfg:
            write_xend_cfg(new_cfg)
    finally:
        lock.release()


def unconfigure_xend(params):
    lock = filelock.LockFile(xend_conf)
    lock.acquire(wait=5)
    try:
        cfg = read_xend_cfg()
        changed = False
        for option in params:
            if option in cfg:
                del cfg[option]
                changed = True
        if changed:
            write_xend_cfg(cfg)
    finally:
        lock.release()


# Begin of copying from yum-config-manager.py
def writeRawConfigFile(filename, sectionname, cfgoptions, items, optionobj, only=None):
    """
    From writeRawRepoFile, but so we can alter [main] too.
    """
    ini = INIConfig(open(filename))
    # Updated the ConfigParser with the changed values
    cfgOptions = cfgoptions(sectionname)
    for name,value in items():
        if value is None: # Proxy
            continue
        option = optionobj(name)
        if option.default != value or name in cfgOptions :
            if only is None or name in only:
                ini[sectionname][name] = option.tostring(value)
    fp =file(filename, "w")
    fp.write(str(ini))
    fp.close()
# End of copy


def get_yum_config():
    lock = filelock.LockFile(yum_conf)
    lock.acquire(wait=5)
    try:
        result = {}
        yb = yum.YumBase()
        result['main'] = {}
        for (option, value) in yb.conf.iteritems():
            result['main'][option] = value
        for repo in sorted(yb.repos.findRepos('*')):
            result[repo.id] = {}
            for (option, value) in repo.iteritems():
                result[repo.id][option] = value
        return result
    finally:
        lock.release()


def configure_yum(section, params):

    def update_conf(conf):
        for (option, value) in params.items():
            if option not in conf.iterkeys():
                raise RuntimeError('Unknown option: %s' % option)
            setattr(conf, option, value)

    lock = filelock.LockFile(yum_conf)
    lock.acquire(wait=5)
    try:
        yb = yum.YumBase()
        if section == 'main':
            ybc = yb.conf
            update_conf(ybc)
            writeRawConfigFile(yum_conf, 'main', ybc.cfg.options, ybc.iteritems, ybc.optionobj, params.keys())
        else:
            try:
                repo = yb.repos.getRepo(section)
            except yum.Errors.RepoError:
                parser = ConfigParser()
                parser.read(yum_repo_conf)
                parser.add_section(section)
                conf = yum.config.RepoConf()
                update_conf(conf)
                for (name, value) in conf.iteritems():
                    if name not in params.keys():
                        continue
                    option = conf.optionobj(name)
                    parser.set(section, name,  option.tostring(value))
                f = open(yum_repo_conf, "w")
                parser.write(f)
                f.close()
                return
            update_conf(repo)
            writeRawConfigFile(repo.repofile, repo.id, repo.cfg.options, repo.iteritems, repo.optionobj, params.keys())
    finally:
        lock.release()


def delete_yum_repo(repo_id):
    lock = filelock.LockFile(yum_conf)
    lock.acquire(wait=5)
    try:
        yb = yum.YumBase()
        try:
            repo = yb.repos.getRepo(repo_id)
        except yum.Errors.RepoError:
            logging.debug('No such repository: %s' % repo_id)
            return
        parser = ConfigParser()
        parser.read(repo.repofile)
        parser.remove_section(repo.id)
        f = open(repo.repofile, "w")
        parser.write(f)
        f.close()
    finally:
        lock.release()


def yum_list_package(pkgnarrow='all', patterns=None, showdups=None, ignore_case=False):

    def format_package_info(pkglist):
        return [(pkg.name, pkg.arch, pkg.epoch, pkg.version, pkg.release, pkg.repo.id) for pkg in pkglist]

    lock = filelock.FileLock(yum_lock)
    lock.acquire(wait=120)
    try:
        yb = yum.YumBase()
        ygh = yb.doPackageLists(pkgnarrow, patterns, showdups, ignore_case)
        result = {}
        for category in ['installed', 'available', 'reinstall_available', 'old_available',
                         'updates', 'obsoletes', 'obsoletesTuples', 'recent', 'extras']:
            result[category] = format_package_info(getattr(ygh, category))
        return result
    finally:
        lock.release()


def yum_update(option=''):
    try:
        run_cmd(['/usr/bin/yum', 'clean', 'all'])
    except Exception, err:
        logging.warn('Cannot remove yum cache date: %s', err)
    cmd = ['/usr/bin/yum', 'update', '-y', '-q']
    if option:
        cmd += [option]
    run_cmd(cmd)


def list_package(name='*'):
    ts = rpm.TransactionSet()
    if re.search('[?*]|\[.+\]', name):
        mi = ts.dbMatch()
        mi.pattern("name", rpm.RPMMIRE_GLOB, name)
    else:
        mi = ts.dbMatch("name", name)
    pkgs = []
    for hdr in mi:
        info = {'name': hdr[rpm.RPMTAG_NAME],
                'arch': hdr[rpm.RPMTAG_ARCH],
                'version': hdr[rpm.RPMTAG_VERSION],
                'release': hdr[rpm.RPMTAG_RELEASE],
                'epoch': hdr[rpm.RPMTAG_EPOCH]}
        pkgs.append(info)
    return pkgs


def install_package(pkg_data, option=''):
    (fd, fn) = tempfile.mkstemp(suffix='.rpm', prefix="rpm-package-")
    try:
        os.write(fd, pkg_data.data)
        os.close(fd)
        cmd = ['/bin/rpm', '--upgrade']
        if option:
            cmd += [option]
        cmd += [fn]
        run_cmd(cmd)
    finally:
        os.unlink(fn)


def uninstall_package(pkg_list, option=''):
    cmd = ['/bin/rpm', '--erase']
    if option:
        cmd += [option]
    cmd += pkg_list
    run_cmd(cmd)


def get_datetime():
    date_backend = dateBackend()
    return tuple(date_backend.getDate())


def set_datetime(year, month, day, hour, min, sec):
    date_backend = dateBackend()
    date_backend.writeDateConfig((year, month - 1, day), (hour, min, sec))
    date_backend.syncHardwareClock()


def get_timezone():
    timezone_backend = timezoneBackend()
    timezone_info = timezone_backend.getTimezoneInfo()
    if timezone_info[1] == 'false':
        return timezone_info[0], False
    else:
        return timezone_info[0], True


def set_timezone(timezone, utc=False):
    if utc:
        utc = 1
    else:
        utc = 0
    timezone_backend = timezoneBackend()
    timezone_backend.writeConfig(timezone, utc)


def get_ntp():
    date_backend = dateBackend()
    (ntp_servers, local_time_source) = date_backend.getNtpServers()
    is_ntp_running = bool(date_backend.isNtpRunning())
    return (ntp_servers, local_time_source, is_ntp_running)


def set_ntp(servers, local_time_source=False, allow_query=False):
    if len(servers) == 0 and local_time_source == False:
        raise Exception('Please specify an NTP server to use or use local time source.')
    logging.debug('substitute manager src ip %s for %s', OVSConstants.MANAGER_SRC_IP, servers)
    servers = [re.sub('0\.0\.0\.0', OVSConstants.MANAGER_SRC_IP, s) for s in servers]
    date_backend = dateBackend()
    broadcast_client = date_backend.getNtpBroadcastClient()
    iburst = date_backend.getIburst()
    if date_backend.writeNtpConfig(servers, broadcast_client, local_time_source, iburst) == None:
        raise Exception('Could not write NTP configurations.')
    fd = open('/etc/sysconfig/ntpd', 'r+')
    lines = fd.readlines()
    fd.seek(0)
    fd.truncate(0)
    for line in lines:
        if re.match(r'^OPTIONS=.*', line) and not re.search(r'("| )-x( |")', line):
            line = re.sub('"$', ' -x"', line)
        fd.write(line)
    fd.close()
    fd = open('/etc/ntp.conf', 'w')
    for line in date_backend.ntpFile:
        if allow_query:
            line = re.sub('^restrict default kod nomodify notrap nopeer noquery$',
                          'restrict default kod nomodify notrap nopeer', line)
            line = re.sub('^restrict -6 default kod nomodify notrap nopeer noquery$',
                          'restrict -6 default kod nomodify notrap nopeer', line)
        else:
            line = re.sub('^restrict default kod nomodify notrap nopeer$',
                          'restrict default kod nomodify notrap nopeer noquery', line)
            line = re.sub('^restrict -6 default kod nomodify notrap nopeer$',
                          'restrict -6 default kod nomodify notrap nopeer noquery', line)
        fd.write(line)
    fd.close()


def enable_ntp():
    date_backend = dateBackend()
    date_backend.startNtpService(None)
    date_backend.chkconfigOn()


def disable_ntp():
    date_backend = dateBackend()
    date_backend.stopNtpService()
    date_backend.chkconfigOff()


def get_ovs_version():
    ovs_version = {}
    f = open("/etc/ovs-info")
    for line in f:
        m = re.match(r'^release:(.*)$', line)
        if m:
            ovs_version['release'] = m.group(1).strip()
        m = re.match(r'^date:(.*)$', line)
        if m:
            ovs_version['date'] = m.group(1).strip()
        m = re.match(r'^build:(.*)$', line)
        if m:
            ovs_version['build'] = m.group(1).strip()
    f.close()
    return ovs_version


def get_log(loglist=None):
    logs = {'messages': '/var/log/messages',
            'dmesg': '/var/log/dmesg',
            'xend': '/var/log/xen/xend.log',
            'ovs-agent': '/var/log/ovs-agent.log'}
    if loglist:
        if not set(loglist) <= set(logs):
            raise Exception('log list contains items not in %s: %s' % (logs.keys(), loglist))
        logs = dict([(k, v) for (k, v) in logs.items() if k in loglist])
    result = {}
    for (logname, filename) in logs.items():
        if os.path.exists(filename):
            result[logname] = ''
            for i in (5, 4, 3, 2, 1):
                rotate_filename = '%s.%s' % (filename, i)
                if os.path.exists(rotate_filename):
                    fd = open(rotate_filename)
                    try:
                        result[logname] += fd.read()
                    finally:
                        fd.close()
            fd = open(filename)
            try:
                result[logname] += fd.read()
            finally:
                fd.close()
            result[logname] = xmlrpclib.Binary(result[logname])
    return result


def get_support_files():
    name = ''.join(random.sample(string.ascii_lowercase, 6))
    for filename in glob.glob('/tmp/sosreport-%s-*.tar.bz2' % name):
        os.unlink(filename)
    cmd = ['/usr/sbin/sosreport', '--batch', '--name=%s' % name]
    run_cmd(cmd)
    filenames = glob.glob('/tmp/sosreport-%s-*.tar.bz2' % name)
    if not filenames:
        raise RuntimeError('No sosreport is generated.')
    filename = filenames[0]
    fd = open(filename)
    data = xmlrpclib.Binary(fd.read())
    fd.close()
    os.unlink(filename)
    os.unlink('%s.md5' % filename)
    return data


def update_agent_password(username, password):
    update_password(username, password)
    reload_agent_server()


def update_core_api_bindings(url, option=None):
    logging.debug('url: %s option: %s', url, option)
    new_url = re.sub("0\.0\.0\.0(?=(:|/))", OVSConstants.MANAGER_SRC_IP, url, 1)
    logging.debug('new url: %s', new_url)
    filename = '/tmp/tmp-%s' % ''.join(random.sample(string.ascii_letters, 6))
    filename = ovs_download_file(new_url, filename, option)
    try:
        cmd = ['unzip', '-o', filename, '-d', get_python_lib()]
        run_cmd(cmd)
        reload_notification_server()
        logging.debug('new core api version: %s', get_core_api_version())
    finally:
        os.unlink(filename)


def get_core_api_version():
    try:
        f = open(os.path.join(get_python_lib(), 'version.properties'))
        d = {}
        for line in f:
            if '=' not in line:
                continue
            (k, v) = line.split('=', 1)
            k = k.strip()
            v = v.strip()
            d[k] = v
        f.close()
        return '.'.join([d['major.version'], d['minor.version'],
                         d['patch.version'], d['bld.version']])
    except Exception, err:
        logging.warn('cannot get core api version: %s', err)
