# Copyright (C) 2006-2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import os
import re
import time
import stat
import errno
import signal
import logging
import subprocess
import threading
from OVSCommons import run_cmd


def guess_mimetype(filename):
    ftype = run_cmd(['file', '-b', '-i', filename])
    ftype = ftype.split(',')[0] # application/x-tar, POSIX
    ftype = ftype.split(';')[0] # application/x-tar; charset=binary
    return ftype.strip()


def set_nice(increment=10, io_cls=2, io_data=7):
    os.nice(increment)
    run_cmd(['ionice', '-c', io_cls, '-n', io_data, '-p', os.getpid()])


def copy_file(src, dst, sparse=True, update_period=20):
    set_nice()
    args = ['/bin/dd', 'if=%s' % src, 'of=%s' % dst, 'bs=1M', 'iflag=direct', 'oflag=direct']
    if sparse:
        args.append('conv=sparse')

    src_stat = os.stat(src)
    src_size = 0
    if stat.S_ISREG(src_stat[stat.ST_MODE]):
        src_size = src_stat[stat.ST_SIZE]
    elif stat.S_ISBLK(src_stat[stat.ST_MODE]):
        src_size = run_cmd(['/sbin/blockdev', '--getsize64', src])

    proc = subprocess.Popen(args, stderr=subprocess.PIPE, close_fds=True)

    def progress_reporter():

        from OVSAsyncProc import async_proc_prog

        try:
            logging.debug('progress reporter for (%s) started', ' '.join(args))
            dd_stat_re = re.compile(r'^(?P<copied>\d+) bytes? (.*) copied.*$')
            time.sleep(update_period)
            proc.poll()
            while proc.returncode is None:
                try:
                    os.kill(proc.pid, signal.SIGUSR1)
                except OSError, err:
                    if err[0] == errno.ESRCH:
                        break
                    else:
                        raise
                while True:
                    dd_stat = proc.stderr.readline()
                    m = dd_stat_re.match(dd_stat)
                    if m:
                        copied = m.group('copied')
                        status = {'OPERATION': 'copy',
                                  'CURRENT_BYTES': long(copied),
                                  'TOTAL_BYTES': long(src_size),
                                  'OBJECT': 'file',
                                  'CURRENT_OBJECT_COUNT': 1,
                                  'TOTAL_OBJECT_COUNT': 1}
                        async_proc_prog(status)
                        break
                time.sleep(update_period)
                proc.poll()
            logging.debug('progress reporter for (%s) finished', ' '.join(args))
        except Exception, err:
            logging.debug('progress reporter for (%s) error: %s', ' '.join(args), err)

    threading.Thread(target=progress_reporter).start()

    proc.wait()
    if proc.returncode != 0:
        raise RuntimeError('Command: (%s) failed: (%s)' % (' '.join(args), proc.returncode))
