# Copyright (C) 2006-2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import os
import time
import fcntl
import errno


class LockError(Exception):
    pass


class LockFile(object):
    def __init__(self, filename):
        self.filename = filename
        self.is_locked = False

    def acquire(self, shared=False, wait=None, delay=0.05):
        if shared:
            op = fcntl.LOCK_SH
        else:
            op = fcntl.LOCK_EX

        if wait is not None:
            op = op | fcntl.LOCK_NB

        try:
            self.file = open(self.filename)
        except IOError, e:
            raise LockError("Open file: %s failed: %s" % (self.filename, e))

        start_time = time.time()
        while True:
            try:
                fcntl.flock(self.file, op)
                break
            except IOError, e:
                if e[0] not in [errno.EACCES, errno.EAGAIN]:
                    raise LockError("Lock file %s failed: %s" % (self.filename, e))
                if (time.time() - start_time) >= wait:
                    raise LockError("Lock file %s failed: timeout occured." % self.filename)
                time.sleep(delay)
        self.is_locked = True

    def release(self):
        if self.is_locked:
            try:
                fcntl.flock(self.file, fcntl.LOCK_UN)
            except IOError, e:
                raise LockError("Unlock file %s failed: %s" % (self.filename, e))
            self.file.close()
            self.is_locked = False

    def __del__(self):
        self.release()


class FileLock(object):
    def __init__(self, filename):
        self.filename = filename
        self.is_locked = False

    def acquire(self, wait=None, delay=0.05):
        start_time = time.time()
        while True:
            try:
                self.fd = os.open(self.filename, os.O_CREAT|os.O_EXCL|os.O_RDWR)
                break
            except OSError, e:
                if wait is None or e[0] != errno.EEXIST:
                    raise LockError("Lock file %s failed: %s" % (self.filename, e))
                if (time.time() - start_time) >= wait:
                    raise LockError("Lock file %s failed: timeout occured." % self.filename)
                time.sleep(delay)
        self.is_locked = True

    def release(self):
        if self.is_locked:
            try:
                os.close(self.fd)
                os.unlink(self.filename)
            except OSError, e:
                raise LockError("Unlock file %s failed: %s" % (self.filename, e))
            self.is_locked = False

    def __del__(self):
        self.release()
